import copy
import freeOrionAIInterface as fo
from collections import Counter
from collections import OrderedDict as odict
from logging import debug, error, info, warning
from operator import itemgetter
from typing import Optional

import AIFleetMission
import ColonisationAI
import CombatRatingsAI
import ExplorationAI
import FleetUtilsAI
import MilitaryAI
import PlanetUtilsAI
from AIDependencies import INVALID_ID, TECH_NATIVE_SPECIALS
from character.character_module import Aggression, create_character
from CombatRatingsAI import ShipCombatStats
from common.print_utils import Number, Table, Text, dict_to_table
from DiplomaticCorp import get_diplomatic_status
from EnumsAI import MissionType, ShipRoleType
from freeorion_tools import combine_ratings, get_partial_visibility_turn

# moving ALL or NEARLY ALL 'global' variables into AIState object rather than module
# in general, leaving items as a module attribute if they are recalculated each turn without reference to prior values
# global variables
from universe.system_network import get_neighbors

colonyTargetedSystemIDs = []
outpostTargetedSystemIDs = []
opponentPlanetIDs = []
invasionTargets = []
invasionTargetedSystemIDs = []
fleetsLostBySystem = {}  # keys are system_ids, values are ratings for the fleets lost
empireStars = {}


class ConversionError(Exception):
    """Exception to be raised if the conversion of a savegame state fails.

    Automatically logs and chats to the host if raised.
    """

    def __init__(self, msg=""):
        error(msg, exc_info=True)


def convert_to_version(state, version):
    """Convert a savegame AIstate to the next version.

    :param dict state: savegame state, modified in function
    :param int version: Version to convert to
    """
    debug("Trying to convert savegame state to version %d..." % version)
    current_version = state.get("version", -1)
    debug("  Current version: %d" % current_version)
    if current_version == version:
        raise ConversionError("Can't convert AI savegame to the same compatibility version.")

    if current_version > version:
        raise ConversionError("Can't convert AI savegame to an older compatibility version.")

    if version != current_version + 1:
        raise ConversionError("Can't skip a compatibility version when converting AI savegame.")

    # Starting with version 3, we switched from pickle to json-style encoding
    # Do not try to load an older savegame even if it magically passed the encoder.
    if version <= 3:
        raise ConversionError("The AI savegame version is no longer supported.")

    if version == 4:
        del state["qualifyingOutpostBaseTargets"]
        del state["qualifyingColonyBaseTargets"]
        state["orbital_colonization_manager"] = ColonisationAI.OrbitalColonizationManager()

    if version == 5:
        state["last_turn_played"] = 0

    if version == 6:
        # Anti-fighter and anti-planet stats were added to CombatRatingAI
        state["_AIstate__empire_standard_enemy"] = state["_AIstate__empire_standard_enemy"] + (0, False) + (0, False)

    if version == 7:
        # This attribute now contains ShipCombatStats object
        standard_enemy = state["_AIstate__empire_standard_enemy"]

        # Changes were done without update so it is possible that users
        # have saved game with state 6 but with new object.
        # So we add addition check that object was not already converted.
        if not isinstance(standard_enemy, ShipCombatStats):
            # Old value was a tuple of arguments
            (
                attacks,
                structure,
                shields,
                fighter_capacity,
                fighter_launch_rate,
                fighter_damage,
                flak_shots,
                has_interceptors,
                damage_vs_planets,
                has_bomber,
            ) = state["_AIstate__empire_standard_enemy"]

            state["_AIstate__empire_standard_enemy"] = ShipCombatStats(
                attacks=attacks,
                structure=structure,
                shields=shields,
                fighter_capacity=fighter_capacity,
                fighter_launch_rate=fighter_launch_rate,
                flak_shots=flak_shots,
                has_interceptors=has_interceptors,
                damage_vs_planets=damage_vs_planets,
                has_bomber=has_bomber,
            )

    #   state["some_new_member"] = some_default_value
    #   del state["some_removed_member"]
    #   state["list_changed_to_set"] = set(state["list_changed_to_set"])

    debug("  All updates set. Setting new version number.")
    state["version"] = version


class AIstate:
    """Stores AI game state.

    IMPORTANT:
    (i) If class members are redefined, added or deleted, then the
    version number must be increased by 1 and the convert_to_version()
    function must be updated so a saved state from the previous version
    is playable with this AIstate version, i.e. new members must be added
    and outdated members must be modified and / or deleted.

    (ii) The AIstate is stored as an encoded string in save game files
    (currently via the pickle module). The attributes of the AIstate must
    therefore be compatible with the encoding method, which currently generally
    means that they must be native python data types (or other data types the
    encoder is augmented to handle), not objects such as UniverseObject
    instances or C++ enum values brought over from the C++ side
    via boost. If desiring to store a reference to a UniverseObject store its
    object id instead; for enum values store their int conversion value.
    """

    version = 7

    def __init__(self, aggression):
        # Do not allow to create AIstate instances with an invalid version number.
        if not hasattr(AIstate, "version"):
            raise ConversionError("AIstate must have an integer version attribute for savegame compatibility")
        if not isinstance(AIstate.version, int):
            raise ConversionError("Version attribute of AIstate must be an integer!")
        if AIstate.version < 0:
            raise ConversionError("AIstate savegame compatibility version must be a positive integer!")

        # need to store the version explicitly as the class variable "version" is only stored in the
        # self.__class__.__dict__ while we only pickle the object (i.e. self.__dict__ )
        self.version = AIstate.version

        # see AIstate docstring re importance of int cast for aggression
        self._aggression = int(aggression)

        # 'global' (?) variables
        self.colonisablePlanetIDs = odict()
        self.colonisableOutpostIDs = odict()  #
        self.__aiMissionsByFleetID = {}
        self.__shipRoleByDesignID = {}
        self.__fleetRoleByID = {}
        self.diplomatic_logs = {}
        self.__priorityByType = {}

        # initialize home system knowledge
        universe = fo.getUniverse()
        empire = fo.getEmpire()
        self.empireID = empire.empireID
        homeworld = universe.getPlanet(empire.capitalID)
        self.__origin_home_system_id = homeworld.systemID if homeworld else INVALID_ID
        self.visBorderSystemIDs = {self.__origin_home_system_id}
        self.visInteriorSystemIDs = set()
        self.exploredSystemIDs = set()
        self.unexploredSystemIDs = {self.__origin_home_system_id}
        self.fleetStatus = {}  # keys: 'sysID', 'nships', 'rating'
        # systemStatus keys:
        # 'name', 'neighbors' (sysIDs), '2jump_ring' (sysIDs), '3jump_ring', '4jump_ring', 'enemy_ship_count',
        # 'fleetThreat', 'planetThreat', 'monsterThreat' (specifically, immobile nonplanet threat), 'totalThreat',
        # 'localEnemyFleetIDs', 'neighborThreat', 'max_neighbor_threat', 'jump2_threat' (up to 2 jumps away),
        # 'jump3_threat', 'jump4_threat', 'regional_threat', 'myDefenses' (planet rating), 'myfleets',
        # 'myFleetsAccessible'(not just next desitination), 'myFleetRating', 'my_neighbor_rating' (up to 1 jump away),
        # 'my_jump2_rating', 'my_jump3_rating', my_jump4_rating', 'local_fleet_threats',
        # 'regional_fleet_threats' <== these are only for mobile fleet threats
        self.systemStatus = {}
        self.needsEmergencyExploration = []
        self.newlySplitFleets = {}
        self.shipCount = 4
        self.misc = {}  # Keys: "enemies_sighted" (dict[turn: list[fleetIDs]]),
        #                       "observed_empires" (set[enemy empire IDs]),
        #                       "ReassignedFleetMissions" (list[FleetMissions])
        self.orbital_colonization_manager = ColonisationAI.OrbitalColonizationManager()
        self.qualifyingTroopBaseTargets = {}
        # TODO: track on a per-empire basis
        self.__empire_standard_enemy = CombatRatingsAI.default_ship_stats()
        self.empire_standard_enemy_rating = 0  # TODO: track on a per-empire basis
        self.character = create_character(aggression, self.empireID)
        self.last_turn_played = 0

    def __setstate__(self, state):
        try:
            for v in range(state.get("version", -1), AIstate.version):
                convert_to_version(state, v + 1)
        except ConversionError:
            if "_aggression" in state:
                aggression = state["_aggression"]
            else:
                try:
                    aggression = state["character"].get_trait(Aggression).key
                except Exception:
                    error("Could not find the aggression level of the AI, defaulting to typical.", exc_info=True)
                    aggression = fo.aggression.typical
            self.__init__(aggression)
            return

        # build the ordered dict with sorted entries from the (unsorted) dict
        # that is contained in the savegame state.
        for content in ("colonisablePlanetIDs", "colonisableOutpostIDs"):
            sorted_planets = sorted(state[content].items(), key=itemgetter(1), reverse=True)
            state[content] = odict(sorted_planets)

        self.__dict__ = state

    def __refresh(self):
        """Turn start AIstate cleanup/refresh."""
        fleetsLostBySystem.clear()
        invasionTargets[:] = []

    def __border_exploration_update(self):
        universe = fo.getUniverse()
        exploration_center = PlanetUtilsAI.get_capital_sys_id()
        # a bad state probably from an old savegame, or else empire has lost (or almost has)
        if exploration_center == INVALID_ID:
            exploration_center = self.__origin_home_system_id
        ExplorationAI.graph_flags.clear()
        if fo.currentTurn() < 50:
            debug("-------------------------------------------------")
            debug("Border Exploration Update (relative to %s)" % universe.getSystem(exploration_center))
            debug("-------------------------------------------------")
        if self.visBorderSystemIDs == {INVALID_ID}:
            self.visBorderSystemIDs.clear()
            self.visBorderSystemIDs.add(exploration_center)
        for sys_id in list(self.visBorderSystemIDs):  # This set is modified during iteration.
            if fo.currentTurn() < 50:
                debug("Considering border system %s" % universe.getSystem(sys_id))
            ExplorationAI.follow_vis_system_connections(sys_id, exploration_center)
        newly_explored = ExplorationAI.update_explored_systems()
        nametags = []
        for sys_id in newly_explored:
            newsys = universe.getSystem(sys_id)
            # an explored system *should* always be able to be gotten
            nametags.append("ID:%4d -- %-20s" % (sys_id, (newsys and newsys.name) or "name unknown"))
        if newly_explored:
            debug("-------------------------------------------------")
            debug("Newly explored systems:\n%s" % "\n".join(nametags))
            debug("-------------------------------------------------")

    def delete_fleet_info(self, fleet_id):
        if fleet_id in self.__aiMissionsByFleetID:
            del self.__aiMissionsByFleetID[fleet_id]
        if fleet_id in self.fleetStatus:
            del self.fleetStatus[fleet_id]
        if fleet_id in self.__fleetRoleByID:
            del self.__fleetRoleByID[fleet_id]
        for sys_status in self.systemStatus.values():
            for fleet_list in [sys_status.get("myfleets", []), sys_status.get("myFleetsAccessible", [])]:
                if fleet_id in fleet_list:
                    fleet_list.remove(fleet_id)

    def __report_system_threats(self):
        """Print a table with system threats to the logfile."""
        current_turn = fo.currentTurn()
        if current_turn >= 100:
            return
        threat_table = Table(
            Text("System"),
            Text("Vis."),
            Number("Total"),
            Number("by Monsters"),
            Number("by Fleets"),
            Number("by Planets"),
            Number("1 jump away"),
            Number("2 jumps"),
            Number("3 jumps"),
            table_name="System Threat Turn %d" % current_turn,
        )
        universe = fo.getUniverse()
        for sys_id in universe.systemIDs:
            sys_status = self.systemStatus.get(sys_id, {})
            system = universe.getSystem(sys_id)
            threat_table.add_row(
                system,
                "Yes" if sys_status.get("currently_visible", False) else "No",
                sys_status.get("totalThreat", 0),
                sys_status.get("monsterThreat", 0),
                sys_status.get("fleetThreat", 0),
                sys_status.get("planetThreat", 0),
                sys_status.get("neighborThreat", 0.0),
                sys_status.get("jump2_threat", 0.0),
                sys_status.get("jump3_threat", 0.0),
            )
        info(threat_table)

    def __report_system_defenses(self):
        """Print a table with system defenses to the logfile."""
        current_turn = fo.currentTurn()
        if current_turn >= 100:
            return
        defense_table = Table(
            Text("System Defenses"),
            Number("Total"),
            Number("by Planets"),
            Number("by Fleets"),
            Number("Fleets 1 jump away"),
            Number("2 jumps"),
            Number("3 jumps"),
            table_name="System Defenses Turn %d" % current_turn,
        )
        universe = fo.getUniverse()
        for sys_id in universe.systemIDs:
            sys_status = self.systemStatus.get(sys_id, {})
            system = universe.getSystem(sys_id)
            defense_table.add_row(
                system,
                sys_status.get("all_local_defenses", 0.0),
                sys_status.get("mydefenses", {}).get("overall", 0.0),
                sys_status.get("myFleetRating", 0.0),
                sys_status.get("my_neighbor_rating", 0.0),
                sys_status.get("my_jump2_rating", 0.0),
                sys_status.get("my_jump3_rating", 0.0),
            )
        info(defense_table)

    def assess_planet_threat(self, pid, sighting_age=0):
        if sighting_age > 5:
            sighting_age += 1  # play it safe
        universe = fo.getUniverse()
        planet = universe.getPlanet(pid)
        if not planet:
            return {"overall": 0, "attack": 0, "health": 0}
        init_shields = planet.initialMeterValue(fo.meterType.shield)
        next_shields = planet.currentMeterValue(fo.meterType.shield)  # always assumes regen will occur
        max_shields = planet.currentMeterValue(fo.meterType.maxShield)
        init_defense = planet.initialMeterValue(fo.meterType.defense)
        next_defense = planet.currentMeterValue(fo.meterType.defense)  # always assumes regen will occur
        max_defense = planet.currentMeterValue(fo.meterType.maxDefense)
        for special, bonuses in TECH_NATIVE_SPECIALS.items():
            if special in planet.specials and sighting_age > 0:
                shield_bonus = bonuses.get("shields", 0)
                defense_bonus = bonuses.get("defense", 0)
                max_shields = max(max_shields, shield_bonus)
                max_defense = max(max_defense, defense_bonus)
                next_shields, init_shields = max(next_shields, shield_bonus), max(init_shields, shield_bonus)
                next_defense, init_defense = max(next_defense, defense_bonus), max(init_defense, defense_bonus)
        # TODO: get regens from knowledge of possessed tech
        # note the max below is because sometimes the next value will be less than init
        # (e.g. shields just after invasion)
        shield_regen = max(1, next_shields - init_shields)
        defense_regen = max(1, next_defense - init_defense)
        shields = min(max_shields, init_shields + sighting_age * shield_regen)
        defense = min(max_defense, init_defense + sighting_age * defense_regen)
        return {"overall": defense * (defense + shields), "attack": defense, "health": (defense + shields)}

    def assess_enemy_supply(self) -> tuple[dict[int, list[int]], dict[int, list[int]]]:
        """
        Assesses where enemy empires have Supply
        :return: a tuple of 2 dicts, each of which is keyed by system id, and each of which is a list of empire ids
        1st dict -- enemies that actually have supply at this system
        2nd dict -- enemies that have supply within 2 jumps from this system (if they clear obstructions)
        """
        enemy_ids = [_id for _id in fo.allEmpireIDs() if _id != fo.empireID()]
        actual_supply = {}
        near_supply = {}
        for enemy_id in enemy_ids:
            this_enemy = fo.getEmpire(enemy_id)
            if not this_enemy:
                debug("Could not retrieve empire for empire id %d" % enemy_id)  # do not spam chat_error with this
                continue
            for sys_id in this_enemy.fleetSupplyableSystemIDs:
                actual_supply.setdefault(sys_id, []).append(enemy_id)
            for sys_id, supply_val in this_enemy.supplyProjections().items():
                if supply_val >= -2:
                    near_supply.setdefault(sys_id, []).append(enemy_id)
        return actual_supply, near_supply

    def __update_empire_standard_enemy(self):
        """Update the empire's standard enemy.

        The standard enemy is the enemy that is most often seen.
        """
        # TODO: If no current information available, rate against own fighters
        universe = fo.getUniverse()
        empire_id = fo.empireID()

        # assess enemy fleets that may have been momentarily visible (start with dummy entries)
        dummy_stats = CombatRatingsAI.default_ship_stats()
        cur_e_fighters = Counter()  # actual visible enemies
        old_e_fighters = Counter({dummy_stats: 0})  # destroyed enemies TODO: consider seen but out of sight enemies

        for fleet_id in universe.fleetIDs:
            fleet = universe.getFleet(fleet_id)
            if (
                not fleet
                or fleet.empty
                or fleet.ownedBy(empire_id)
                or fleet.unowned
                or not (fleet.hasArmedShips or fleet.hasFighterShips)
            ):
                continue

            # track old/dead enemy fighters for rating assessments in case not enough current info
            ship_stats = CombatRatingsAI.get_ships_stats_for_fleet(fleet_id)
            dead_fleet = fleet_id in universe.destroyedObjectIDs(empire_id)
            e_f_dict = old_e_fighters if dead_fleet else cur_e_fighters
            for stats in ship_stats:
                # log only ships that are armed
                if stats.attacks:
                    e_f_dict[stats] += 1

        e_f_dict = cur_e_fighters or old_e_fighters
        self.__empire_standard_enemy = max(e_f_dict, key=e_f_dict.get)
        self.empire_standard_enemy_rating = self.get_standard_enemy().get_rating()

    def __update_system_status(self):  # noqa: C901
        debug("=== Updating System Threats ===")
        universe = fo.getUniverse()
        empire = fo.getEmpire()
        empire_id = fo.empireID()
        destroyed_object_ids = universe.destroyedObjectIDs(empire_id)
        supply_unobstructed_systems = set(empire.supplyUnobstructedSystems)
        min_hidden_attack = 4
        min_hidden_health = 8
        observed_empires = self.misc.setdefault("observed_empires", set())

        # TODO: Variables that are recalculated each turn from scratch should not be stored in AIstate
        # clear previous game state
        for sys_id in self.systemStatus:
            self.systemStatus[sys_id]["enemy_ship_count"] = 0
            self.systemStatus[sys_id]["myFleetRating"] = 0
            self.systemStatus[sys_id]["myFleetRatingVsPlanets"] = 0

        # for use in debugging
        verbose = False

        # assess enemy fleets that may have been momentarily visible
        enemies_by_system = {}
        my_fleets_by_system = {}
        fleet_spot_position = {}
        current_turn = fo.currentTurn()
        for fleet_id in universe.fleetIDs:
            fleet = universe.getFleet(fleet_id)
            if not fleet or fleet.empty:
                self.delete_fleet_info(fleet_id)  # this is safe even if fleet wasn't mine
                continue
            # TODO: check if currently in system and blockaded before accepting destination as location
            this_system_id = fleet.nextSystemID if fleet.nextSystemID != INVALID_ID else fleet.systemID
            dead_fleet = fleet_id in destroyed_object_ids
            if dead_fleet:
                self.delete_fleet_info(fleet_id)

            if fleet.ownedBy(empire_id):
                if not dead_fleet:
                    my_fleets_by_system.setdefault(this_system_id, []).append(fleet_id)
                    fleet_spot_position.setdefault(fleet.systemID, []).append(fleet_id)
                continue

            # TODO: consider checking death of individual ships.  If ships had been moved from this fleet
            # into another fleet, we might have witnessed their death in that other fleet but if this fleet
            # had not been seen since before that transfer then the ships might also still be listed here.
            if dead_fleet:
                continue

            # we are only interested in immediately recent data
            if get_partial_visibility_turn(fleet_id) < (current_turn - 1):
                continue

            if get_diplomatic_status(fleet.owner) != fo.diplomaticStatus.war:
                continue

            sys_status = self.systemStatus.setdefault(this_system_id, {})
            sys_status["enemy_ship_count"] = sys_status.get("enemy_ship_count", 0) + len(fleet.shipIDs)
            enemies_by_system.setdefault(this_system_id, []).append(fleet_id)

            if not fleet.unowned:
                self.misc.setdefault("enemies_sighted", {}).setdefault(current_turn, []).append(fleet_id)
                observed_empires.add(fleet.owner)

        # assess fleet and planet threats & my local fleets
        for sys_id in universe.systemIDs:
            sys_status = self.systemStatus.setdefault(sys_id, {})
            system = universe.getSystem(sys_id)
            if verbose:
                debug("AIState threat evaluation for %s" % system)
            # update fleets
            sys_status["myfleets"] = my_fleets_by_system.get(sys_id, [])
            sys_status["myFleetsAccessible"] = fleet_spot_position.get(sys_id, [])
            local_enemy_fleet_ids = enemies_by_system.get(sys_id, [])
            sys_status["localEnemyFleetIDs"] = local_enemy_fleet_ids
            if system:
                sys_status["name"] = system.name

            # update my fleet rating versus planets so that planet ratings can be more accurate
            my_ratings_against_planets_list = []
            for fid in sys_status["myfleets"]:
                my_ratings_against_planets_list.append(self.get_rating(fid, against_planets=True))
                sys_status["myFleetRatingVsPlanets"] = combine_ratings(my_ratings_against_planets_list)

            # update threats
            monster_ratings = []  # immobile
            enemy_ratings = []  # owned & mobile
            mob_ratings = []  # mobile & unowned
            mobile_fleets = []  # mobile and either owned or unowned
            for fid in local_enemy_fleet_ids:
                fleet = universe.getFleet(fid)  # ensured to exist
                fleet_rating = CombatRatingsAI.get_fleet_rating(
                    fid, enemy_stats=CombatRatingsAI.get_empire_standard_military_ship_stats()
                )
                if fleet.speed == 0:
                    monster_ratings.append(fleet_rating)
                    if verbose:
                        debug(f"\t immobile enemy fleet {fleet} has rating {fleet_rating:.1f}")
                    continue

                if verbose:
                    debug(f"\t mobile enemy fleet {fleet} has rating {fleet_rating:.1f}")
                mobile_fleets.append(fid)
                if fleet.unowned:
                    mob_ratings.append(fleet_rating)
                else:
                    enemy_ratings.append(fleet_rating)

            enemy_rating = combine_ratings(enemy_ratings)
            monster_rating = combine_ratings(monster_ratings)
            mob_rating = combine_ratings(mob_ratings)
            lost_fleets = fleetsLostBySystem.get(sys_id, [])
            lost_fleet_rating = combine_ratings(lost_fleets)
            if lost_fleet_rating:
                debug("Just lost fleet rating %.1f in system %s", lost_fleet_rating, system)

            # under current visibility rules should not be possible to have any losses or other info here,
            # but just in case...
            partial_vis_turn = get_partial_visibility_turn(sys_id)
            if not system or partial_vis_turn < 0:
                if verbose:
                    debug("Never had partial vis for %s - basing threat assessment on old info and lost ships" % system)
                sys_status.setdefault("local_fleet_threats", set())
                sys_status["planetThreat"] = 0
                sys_status["fleetThreat"] = max(
                    combine_ratings(enemy_rating, mob_rating),
                    0.98 * sys_status.get("fleetThreat", 0),
                    1.1 * lost_fleet_rating - monster_rating,
                )
                sys_status["monsterThreat"] = max(
                    monster_rating,
                    0.98 * sys_status.get("monsterThreat", 0),
                    1.1 * lost_fleet_rating - enemy_rating - mob_rating,
                )
                sys_status["enemy_threat"] = max(
                    enemy_rating,
                    0.98 * sys_status.get("enemy_threat", 0),
                    1.1 * lost_fleet_rating - monster_rating - mob_rating,
                )
                sys_status["mydefenses"] = {"overall": 0, "attack": 0, "health": 0}
                sys_status["totalThreat"] = sys_status["fleetThreat"]
                sys_status["regional_fleet_threats"] = sys_status["local_fleet_threats"].copy()
                continue

            # have either stale or current info
            pattack = phealth = 0
            mypattack = myphealth = 0
            for pid in system.planetIDs:
                planet = universe.getPlanet(pid)
                if not planet:
                    continue
                sighting_age = current_turn - get_partial_visibility_turn(pid)
                prating = self.assess_planet_threat(pid, sighting_age)
                if planet.ownedBy(empire_id):  # TODO: check for diplomatic status
                    mypattack += prating["attack"]
                    myphealth += prating["health"]
                else:
                    pattack += prating["attack"]
                    phealth += prating["health"]
                    if any("_NEST_" in special for special in planet.specials):
                        sys_status["nest_threat"] = 100
            sys_status["planetThreat"] = pattack * phealth
            sys_status["mydefenses"] = {"overall": mypattack * myphealth, "attack": mypattack, "health": myphealth}

            # previous threat assessment could account for losses, ignore the losses now
            if lost_fleet_rating and lost_fleet_rating < max(sys_status.get("totalThreat", 0), pattack * phealth):
                debug("In system %s: Ignoring lost fleets since known threats could cause it.", system)
                lost_fleet_rating = 0

            # TODO also only consider past stealthed fleet threat to still be present if the system is still obstructed
            # TODO: track visibility across turns in order to distinguish the blip of visibility in (losing) combat,
            #       which FO currently treats as being for the previous turn,
            #       partially superseding the previous visibility for that turn

            # Estimating stealthed enemies by fleets lost to them
            if not partial_vis_turn == current_turn:
                sys_status.setdefault("local_fleet_threats", set())
                sys_status["currently_visible"] = False
                # print ("Stale visibility for system %d ( %s ) -- last seen %d, "
                #        "current Turn %d -- basing threat assessment on old info and lost ships") % (
                #     sys_id, sys_status.get('name', "name unknown"), partial_vis_turn, currentTurn)
                sys_status["fleetThreat"] = max(
                    combine_ratings(enemy_rating, mob_rating),
                    0.98 * sys_status.get("fleetThreat", 0),
                    2.0 * lost_fleet_rating - max(sys_status.get("monsterThreat", 0), monster_rating),
                )
                sys_status["enemy_threat"] = max(
                    enemy_rating,
                    0.98 * sys_status.get("enemy_threat", 0),
                    1.1 * lost_fleet_rating - max(sys_status.get("monsterThreat", 0), monster_rating),
                )
                sys_status["monsterThreat"] = max(monster_rating, 0.98 * sys_status.get("monsterThreat", 0))
                # sys_status['totalThreat'] = ((pattack + enemy_attack + monster_attack) ** 0.8)\
                #                             * ((phealth + enemy_health + monster_health)** 0.6)  # reevaluate this
                sys_status["totalThreat"] = max(
                    combine_ratings(enemy_rating, mob_rating, monster_rating, pattack * phealth),
                    2 * lost_fleet_rating,
                    0.98 * sys_status.get("totalThreat", 0),
                )
            else:  # system considered visible
                sys_status["currently_visible"] = True
                sys_status["local_fleet_threats"] = set(mobile_fleets)
                # includes mobile monsters
                sys_status["fleetThreat"] = max(
                    combine_ratings(enemy_rating, mob_rating), 2 * lost_fleet_rating - monster_rating
                )
                if verbose:
                    debug(
                        f"enemy threat calc parts: enemy rating {enemy_rating:.1f}, lost fleet rating {lost_fleet_rating:.1f}, monster_rating {monster_rating:.1f}"
                    )
                # does NOT include mobile monsters
                sys_status["enemy_threat"] = max(enemy_rating, 2 * lost_fleet_rating - monster_rating)
                sys_status["monsterThreat"] = monster_rating
                sys_status["totalThreat"] = combine_ratings(
                    sys_status["fleetThreat"],
                    sys_status["monsterThreat"],
                    pattack * phealth,
                )
            sys_status["regional_fleet_threats"] = sys_status["local_fleet_threats"].copy()
            sys_status["fleetThreat"] = max(sys_status["fleetThreat"], sys_status.get("nest_threat", 0))
            sys_status["totalThreat"] = max(sys_status["totalThreat"], sys_status.get("nest_threat", 0))

            # has been seen with Partial Vis, but is currently supply-blocked
            if partial_vis_turn > 0 and sys_id not in supply_unobstructed_systems:
                sys_status["fleetThreat"] = max(sys_status["fleetThreat"], min_hidden_attack * min_hidden_health)
                sys_status["totalThreat"] = max(
                    sys_status["totalThreat"],
                    combine_ratings(sys_status.get("planetThreat", 0), (min_hidden_attack * min_hidden_health)),
                )
            if verbose and sys_status["fleetThreat"] > 0:
                debug(f"{system} intermediate status: {sys_status}")

        enemy_supply, enemy_near_supply = self.assess_enemy_supply()  # TODO: assess change in enemy supply over time
        # assess secondary threats (threats of surrounding systems) and update my fleet rating
        for sys_id in universe.systemIDs:
            sys_status = self.systemStatus[sys_id]
            sys_status["enemies_supplied"] = enemy_supply.get(sys_id, [])
            observed_empires.update(enemy_supply.get(sys_id, []))
            sys_status["enemies_nearly_supplied"] = enemy_near_supply.get(sys_id, [])
            my_ratings_list = []
            my_ratings_against_planets_list = []
            for fid in sys_status["myfleets"]:
                this_rating = self.get_rating(fid, True, self.get_standard_enemy())
                my_ratings_list.append(this_rating)
                my_ratings_against_planets_list.append(self.get_rating(fid, against_planets=True))
            if sys_id != INVALID_ID:
                sys_status["myFleetRating"] = combine_ratings(my_ratings_list)
                sys_status["myFleetRatingVsPlanets"] = combine_ratings(my_ratings_against_planets_list)
                sys_status["all_local_defenses"] = combine_ratings(
                    sys_status["myFleetRating"], sys_status["mydefenses"]["overall"]
                )
            sys_status["neighbors"] = get_neighbors(sys_id)

        for sys_id in universe.systemIDs:
            sys_status = self.systemStatus[sys_id]
            neighbors = sys_status.get("neighbors", set())
            this_system = universe.getSystem(sys_id)
            if verbose:
                debug(
                    "Regional Assessment for {} with local fleet threat {:.1f}".format(
                        this_system, sys_status.get("fleetThreat", 0)
                    )
                )
            jumps2 = set()
            jumps3 = set()
            jumps4 = set()
            for seta, setb in [(neighbors, jumps2), (jumps2, jumps3), (jumps3, jumps4)]:
                for sys2id in seta:
                    setb.update(self.systemStatus.get(sys2id, {}).get("neighbors", set()))
            jump2ring = jumps2 - neighbors - {sys_id}
            jump3ring = jumps3 - jumps2 - neighbors - {sys_id}
            jump4ring = jumps4 - jumps3 - jumps2 - neighbors - {sys_id}
            sys_status["2jump_ring"] = jump2ring
            sys_status["3jump_ring"] = jump3ring
            sys_status["4jump_ring"] = jump4ring
            threat, max_threat, myrating, j1_threats = self.area_ratings(neighbors)
            sys_status["neighborThreat"] = threat
            sys_status["max_neighbor_threat"] = max_threat
            sys_status["my_neighbor_rating"] = myrating
            threat, max_threat, myrating, j2_threats = self.area_ratings(jump2ring)
            sys_status["jump2_threat"] = threat
            sys_status["my_jump2_rating"] = myrating
            threat, max_threat, myrating, j3_threats = self.area_ratings(jump3ring)
            sys_status["jump3_threat"] = threat
            sys_status["my_jump3_rating"] = myrating
            # for local system includes both enemies and mobs
            threat_keys = ["fleetThreat", "neighborThreat", "jump2_threat"]
            sys_status["regional_threat"] = combine_ratings(sys_status.get(x, 0) for x in threat_keys)
            # TODO: investigate cases where regional_threat has been nonzero but no regional_threat_fleets
            # (probably due to attenuating history of past threats)
            sys_status.setdefault("regional_fleet_threats", set()).update(j1_threats, j2_threats)

    def area_ratings(self, system_ids):
        """Returns (fleet_threat, max_threat, myFleetRating, threat_fleets) compiled over a group of systems."""
        myrating = threat = max_threat = 0
        threat_fleets = set()
        for sys_id in system_ids:
            sys_status = self.systemStatus.get(sys_id, {})
            # TODO: have distinct treatment for both enemy_threat and fleetThreat, respectively
            fthreat = sys_status.get("enemy_threat", 0)
            max_threat = max(max_threat, fthreat)
            threat = combine_ratings(threat, fthreat)
            myrating = combine_ratings(myrating, sys_status.get("myFleetRating", 0))
            # myrating = FleetUtilsAI.combine_ratings(myrating, sys_status.get('all_local_defenses', 0))
            threat_fleets.update(sys_status.get("local_fleet_threats", []))
        return threat, max_threat, myrating, threat_fleets

    def get_fleet_mission(self, fleet_id: int) -> Optional[AIFleetMission.AIFleetMission]:
        """
        Returns AIFleetMission with fleetID.
        """
        if fleet_id in self.__aiMissionsByFleetID:
            return self.__aiMissionsByFleetID[fleet_id]
        else:
            return None

    def get_all_fleet_missions(self):
        """Returns all AIFleetMissions."""
        return self.__aiMissionsByFleetID.values()

    def get_fleet_missions_map(self):
        return self.__aiMissionsByFleetID

    def get_fleet_missions_with_any_mission_types(self, mission_types):
        """Returns all AIFleetMissions which contains any of fleetMissionTypes."""
        result = []
        for mission in self.get_all_fleet_missions():
            if mission.type in mission_types:
                result.append(mission)
        return result

    def __add_fleet_mission(self, fleet_id):
        """Add a new dummy AIFleetMission for the passed fleet_id if it has no mission yet."""
        if self.get_fleet_mission(fleet_id) is not None:
            warning("Tried to add a new fleet mission for fleet that already had a mission.")
            return
        self.__aiMissionsByFleetID[fleet_id] = AIFleetMission.AIFleetMission(fleet_id)

    def __remove_fleet_mission(self, fleet_id):
        """Remove invalid AIFleetMission with fleetID if it exists."""
        if self.get_fleet_mission(fleet_id) is not None:
            self.__aiMissionsByFleetID[fleet_id] = None
            del self.__aiMissionsByFleetID[fleet_id]

    def ensure_have_fleet_missions(self, fleet_ids):
        for fleet_id in fleet_ids:
            if self.get_fleet_mission(fleet_id) is None:
                self.__add_fleet_mission(fleet_id)

    def __clean_fleet_missions(self):
        """Assign a new dummy mission to new fleets and clean up existing, now invalid missions."""
        current_empire_fleets = FleetUtilsAI.get_empire_fleet_ids()

        # assign a new (dummy) mission to new fleets
        for fleet_id in current_empire_fleets:
            if self.get_fleet_mission(fleet_id) is None:
                self.__add_fleet_mission(fleet_id)

        # Check all fleet missions for validity and clear invalid targets.
        # If a fleet does not exist anymore, mark mission for deletion.
        # Deleting only after the loop allows us to avoid an expensive copy.
        deleted_fleet_ids = []
        for mission in self.get_all_fleet_missions():
            if mission.fleet.id not in current_empire_fleets:
                deleted_fleet_ids.append(mission.fleet.id)
            else:
                mission.clean_invalid_targets()
        for deleted_fleet_id in deleted_fleet_ids:
            self.__remove_fleet_mission(deleted_fleet_id)

    def has_target(self, mission_type, target):
        for mission in self.get_fleet_missions_with_any_mission_types([mission_type]):
            if mission.has_target(mission_type, target):
                return True
        return False

    def get_rating(self, fleet_id, force_new=False, enemy_stats=None, against_planets=False):
        """Returns a dict with various rating info."""
        if fleet_id in self.fleetStatus and not force_new and enemy_stats is None:
            return self.fleetStatus[fleet_id].get("rating", 0)
        else:
            fleet = fo.getUniverse().getFleet(fleet_id)
            if not fleet:
                return {}  # TODO: also ensure any info for that fleet is deleted
            status = {
                "rating": CombatRatingsAI.get_fleet_rating(fleet_id, enemy_stats),
                "ratingVsPlanets": CombatRatingsAI.get_fleet_rating_against_planets(fleet_id),
                "sysID": fleet.systemID,
                "nships": len(fleet.shipIDs),
            }
            self.fleetStatus[fleet_id] = status
            return status["rating"] if not against_planets else status["ratingVsPlanets"]

    def update_fleet_rating(self, fleet_id):
        self.get_rating(fleet_id, force_new=True)

    def get_ship_role(self, ship_design_id):
        """Returns ship role for given designID, assesses and adds as needed."""

        # if thought was invalid, recheck to be sure
        if (
            ship_design_id in self.__shipRoleByDesignID
            and self.__shipRoleByDesignID[ship_design_id] != ShipRoleType.INVALID
        ):
            return self.__shipRoleByDesignID[ship_design_id]
        else:
            role = FleetUtilsAI.assess_ship_design_role(fo.getShipDesign(ship_design_id))
            self.__shipRoleByDesignID[ship_design_id] = role
            return role

    def get_fleet_roles_map(self):
        return self.__fleetRoleByID

    def get_fleet_role(self, fleet_id, force_new=False):
        """Returns fleet role by ID."""

        if not force_new and fleet_id in self.__fleetRoleByID:
            return self.__fleetRoleByID[fleet_id]
        else:
            role = FleetUtilsAI.assess_fleet_role(fleet_id)
            self.__fleetRoleByID[fleet_id] = role
            make_aggressive = False
            if role in [
                MissionType.COLONISATION,
                MissionType.OUTPOST,
                MissionType.ORBITAL_INVASION,
                MissionType.ORBITAL_OUTPOST,
            ]:
                pass
            elif role in [MissionType.EXPLORATION, MissionType.INVASION]:
                this_rating = self.get_rating(fleet_id)  # Done!
                n_ships = self.fleetStatus.get(fleet_id, {}).get("nships", 1)  # entry sould exist due to above line
                if float(this_rating) / n_ships >= 0.5 * MilitaryAI.cur_best_mil_ship_rating():
                    make_aggressive = True
            else:
                make_aggressive = True
            fo.issueAggressionOrder(fleet_id, make_aggressive)
            return role

    def session_start_cleanup(self):
        self.newlySplitFleets = {}
        for fleetID in FleetUtilsAI.get_empire_fleet_ids():
            self.get_fleet_role(fleetID)
            self.update_fleet_rating(fleetID)
            self.ensure_have_fleet_missions([fleetID])
        self.__clean_fleet_roles(just_resumed=True)
        fleetsLostBySystem.clear()
        empireStars.clear()
        self.qualifyingTroopBaseTargets.clear()

    def __clean_fleet_roles(self, just_resumed=False):
        """Removes fleetRoles if a fleet has been lost, and update fleet Ratings."""
        universe = fo.getUniverse()
        current_empire_fleets = FleetUtilsAI.get_empire_fleet_ids()
        self.shipCount = 0

        fleet_table = Table(
            Text("Fleet"),
            Number("Rating"),
            Number("Troops"),
            Text("Location"),
            Text("Destination"),
            table_name="Fleet Summary Turn %d" % fo.currentTurn(),
        )
        # need to loop over a copy as entries are deleted in loop
        for fleet_id in list(self.__fleetRoleByID):
            fleet_status = self.fleetStatus.setdefault(fleet_id, {})
            rating = CombatRatingsAI.get_fleet_rating(fleet_id)
            old_sys_id = fleet_status.get("sysID", -2)  # TODO: Introduce helper function instead
            fleet = universe.getFleet(fleet_id)
            if fleet:
                sys_id = fleet.systemID
                if old_sys_id in [-2, -1]:
                    old_sys_id = sys_id
                fleet_status["nships"] = len(fleet.shipIDs)  # TODO: Introduce helper function instead
                self.shipCount += fleet_status["nships"]
            else:
                # can still retrieve a fleet object even if fleet was just destroyed, so shouldn't get here
                # however,this has been observed happening, and is the reason a fleet check was added a few lines below.
                # Not at all sure how this came about, but was throwing off threat assessments
                sys_id = old_sys_id

            # check if fleet is destroyed and if so, delete stored information
            if fleet_id not in current_empire_fleets:  # or fleet.empty:
                debug("Just lost %s", fleet)
                if not just_resumed:
                    fleetsLostBySystem.setdefault(old_sys_id, []).append(
                        max(rating, fleet_status.get("rating", 0.0), MilitaryAI.MinThreat)
                    )

                self.delete_fleet_info(fleet_id)
                continue

            # if reached here, the fleet does still exist
            this_sys = universe.getSystem(sys_id)
            next_sys = universe.getSystem(fleet.nextSystemID)

            fleet_table.add_row(
                fleet,
                rating,
                FleetUtilsAI.count_troops_in_fleet(fleet_id),
                this_sys or "starlane",
                next_sys or "-",
            )

            fleet_status["rating"] = rating
            if next_sys:
                fleet_status["sysID"] = next_sys.id
            elif this_sys:
                fleet_status["sysID"] = this_sys.id
            else:
                warning("Fleet %s has no valid system." % fleet)
        info(fleet_table)
        debug(
            dict_to_table(
                CombatRatingsAI.get_empire_standard_military_ship_stats().__getstate__(),
                name="Empire standard fighter summary",
            )
        )

    def get_explored_system_ids(self):
        return list(self.exploredSystemIDs)

    def get_unexplored_system_ids(self):
        return list(self.unexploredSystemIDs)

    def set_priority(self, priority_type, value):
        """Sets a priority of the specified type."""
        self.__priorityByType[priority_type] = value

    def get_priority(self, priority_type):
        """Returns the priority value of the specified type."""

        if priority_type in self.__priorityByType:
            return copy.deepcopy(self.__priorityByType[priority_type])
        return 0

    def __report_last_turn_fleet_missions(self):
        """Print a table reviewing last turn fleet missions to the log file."""
        universe = fo.getUniverse()
        mission_table = Table(
            Text("Fleet"),
            Text("Mission"),
            Text("Ships"),
            Number("Rating"),
            Number("Troops"),
            Text("Target"),
            table_name="Turn %d: Fleet Mission Review from Last Turn" % fo.currentTurn(),
        )
        for fleet_id, mission in self.get_fleet_missions_map().items():
            fleet = universe.getFleet(fleet_id)
            if not fleet:
                continue
            if not mission:
                mission_table.add_row(fleet)
            else:
                mission_table.add_row(
                    fleet,
                    mission.type or "None",
                    len(fleet.shipIDs),
                    CombatRatingsAI.get_fleet_rating(fleet_id),
                    FleetUtilsAI.count_troops_in_fleet(fleet_id),
                    mission.target or "-",
                )
        info(mission_table)

    def __split_new_fleets(self):
        """Split any new fleets.

        This function is supposed to be called once at the beginning of the turn.
        Splitting the auto generated fleets at game start or those created by
        recently built ships allows the AI to assign correct roles to all ships.
        """
        # TODO: check length of fleets for losses or do in AIstate.__cleanRoles
        universe = fo.getUniverse()
        known_fleets = self.get_fleet_roles_map()
        self.newlySplitFleets.clear()

        fleets_to_split = [fleet_id for fleet_id in FleetUtilsAI.get_empire_fleet_ids() if fleet_id not in known_fleets]
        if fleets_to_split:
            debug("Trying to split %d new fleets" % len(fleets_to_split))
        for fleet_id in fleets_to_split:
            fleet = universe.getFleet(fleet_id)
            if not fleet:
                warning("Trying to split fleet %d but seemingly does not exist" % fleet_id)
                continue
            fleet_len = len(fleet.shipIDs)
            if fleet_len == 1:
                continue
            new_fleets = FleetUtilsAI.split_fleet(fleet_id)
            debug("Split fleet %d with %d ships into %d new fleets:" % (fleet_id, fleet_len, len(new_fleets)))
            # old fleet may have different role after split, later will be again identified
            # in current system, orig new fleet will not yet have been assigned a role
            # self.remove_fleet_role(fleet_id)

    def __cleanup_qualifiying_base_targets(self):
        """Cleanup invalid entries in qualifying base targets."""
        universe = fo.getUniverse()
        empire_id = fo.empireID()
        for dct in [self.qualifyingTroopBaseTargets]:
            for pid in list(dct.keys()):
                planet = universe.getPlanet(pid)
                if planet and planet.ownedBy(empire_id):
                    del dct[pid]

    def prepare_for_new_turn(self):
        self.__report_last_turn_fleet_missions()
        self.__split_new_fleets()
        self.__refresh()  # TODO: Use turn_state instead
        self.__border_exploration_update()
        self.__cleanup_qualifiying_base_targets()
        self.orbital_colonization_manager.turn_start_cleanup()
        self.__clean_fleet_roles()
        self.__clean_fleet_missions()
        debug("Fleets lost by system: %s" % fleetsLostBySystem)
        self.__update_empire_standard_enemy()
        self.__update_system_status()
        self.__report_system_threats()
        self.__report_system_defenses()
        self.__report_exploration_status()

    def __report_exploration_status(self):
        universe = fo.getUniverse()
        explored_system_ids = self.get_explored_system_ids()
        debug("Unexplored Systems: %s " % [universe.getSystem(sys_id) for sys_id in self.get_unexplored_system_ids()])
        debug("Explored SystemIDs: %s" % [universe.getSystem(sys_id) for sys_id in explored_system_ids])
        debug("Explored PlanetIDs: %s" % PlanetUtilsAI.get_planets_in__systems_ids(explored_system_ids))

    def log_alliance_request(self, initiating_empire_id, recipient_empire_id):
        """Keep a record of alliance requests made or received by this empire."""

        alliance_requests = self.diplomatic_logs.setdefault("alliance_requests", {})
        log_index = (initiating_empire_id, recipient_empire_id)
        alliance_requests.setdefault(log_index, []).append(fo.currentTurn())

    def log_peace_request(self, initiating_empire_id, recipient_empire_id):
        """Keep a record of peace requests made or received by this empire."""

        peace_requests = self.diplomatic_logs.setdefault("peace_requests", {})
        log_index = (initiating_empire_id, recipient_empire_id)
        peace_requests.setdefault(log_index, []).append(fo.currentTurn())

    def log_war_declaration(self, initiating_empire_id, recipient_empire_id):
        """Keep a record of war declarations made or received by this empire."""

        # if war declaration is made on turn 1, don't hold it against them
        if fo.currentTurn() == 1:
            return
        war_declarations = self.diplomatic_logs.setdefault("war_declarations", {})
        log_index = (initiating_empire_id, recipient_empire_id)
        war_declarations.setdefault(log_index, []).append(fo.currentTurn())

    def get_standard_enemy(self):
        return self.__empire_standard_enemy



from focs._effects import (
    DirectDistanceBetween,
    EffectsGroup,
    IsField,
    IsSource,
    LocalCandidate,
    MinimumNumberOf,
    MoveTowards,
    Source,
)
from focs._species import *
from species.species_macros.detection import GREAT_DETECTION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import DESERT_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import NO_INDUSTRY
from species.species_macros.influence import BAD_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import BAD_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import GOOD_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import BAD_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS

Species(
    name="SP_NYMNMN",
    description="SP_NYMNMN_DESC",
    gameplay_description="SP_NYMNMN_GAMEPLAY_DESC",
    native=True,
    tags=[
        "SELF_SUSTAINING",
        "NO_INDUSTRY",
        "BAD_INFLUENCE",
        "BAD_POPULATION",
        "BAD_SUPPLY",
        "GREAT_DETECTION",
        "GOOD_STEALTH",
        "PEDIA_SELF_SUSTAINING_SPECIES_CLASS",
    ],
    foci=[
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "BLD_NEUTRONIUM_EXTRACTOR",
        "BLD_NEUTRONIUM_FORGE",
        "BLD_NEUTRONIUM_SYNTH",
        "MINERALS_SPECIAL",
        "MONOPOLE_SPECIAL",
        "POSITRONIUM_SPECIAL",
    ],
    dislikes=[
        "BLD_GAS_GIANT_GEN",
        "BLD_SOL_ORB_GEN",
        "BLD_SPATIAL_DISTORT_GEN",
        "ECCENTRIC_ORBIT_SPECIAL",
        "FORTRESS_SPECIAL",
        "FRACTAL_GEODES_SPECIAL",
        "GAIA_SPECIAL",
        "PLC_DESIGN_SIMPLICITY",
        "SPARK_FOSSILS_SPECIAL",
    ],
    effectsgroups=[
        NO_INDUSTRY,
        *AVERAGE_RESEARCH,
        *BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *BAD_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_NYMNMN"),
        *BAD_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *GREAT_DETECTION,
        *GOOD_STEALTH,
        EffectsGroup(
            scope=MinimumNumberOf(
                number=1,
                sortkey=DirectDistanceBetween(Source.ID, LocalCandidate.ID),
                condition=IsField(name=["FLD_ION_STORM", "FLD_METEOR_BLIZZARD", "FLD_NANITE_SWARM"]),
            ),
            effects=MoveTowards(speed=5, target=IsSource),
        ),
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=DESERT_STANDARD_EP,
    graphic="icons/species/intangible-05.png",
)



import freeOrionAIInterface as fo
import inspect
import pprint
import re
import traceback
from collections.abc import Mapping
from functools import wraps
from logging import ERROR, Handler, debug, error, getLogger, warning
from typing import Optional

import AIDependencies
from common.configure_logging import FOLogFormatter
from common.fo_typing import SpeciesName
from freeorion_tools.caching import cache_for_current_turn, cache_for_session

# color wrappers for chat:
RED = "<rgba 255 0 0 255>%s</rgba>"
WHITE = "<rgba 255 255 255 255>%s</rgba>"


def dict_from_map(thismap):
    """Convert C++ map to python dict."""
    return {el.key(): el.data() for el in thismap}


def dict_from_map_recursive(thismap):
    retval = {}
    try:
        for el in thismap:
            retval[el.key()] = dict_from_map_recursive(el.data())
        return retval
    except Exception:
        return dict_from_map(thismap)


def get_ai_tag_grade(tag_list, tag_type):
    """
    Accepts a list of string tags and a tag_type (like 'WEAPONS').
    Checks for the first tag in the list (if any), for tag_type "TYPE",
    having the structure X_TYPE
    and then returns 'X'
    X is most commonly (but not necessarily) one of [NO, VERY_BAD, BAD, AVERAGE, GOOD, GREAT, ULTIMATE]
    If no matching tags, returns empty string (which for most types should be considered equivalent to AVERAGE)
    """
    for tag in [tag_ for tag_ in tag_list if tag_.count("_") > 0]:
        parts = tag.split("_", 1)
        if parts[1].startswith("BAD_"):
            parts = [parts[0] + "_BAD", parts[1][4:]]
        if parts[1] == tag_type.upper():
            return parts[0]
    return ""


def tech_is_complete(tech: str) -> bool:
    """
    Return if tech has been researched.
    """
    return fo.getEmpire().techResearched(tech)


def tech_soon_available(tech: str, max_position: int) -> bool:
    """
    Return if tech has been researched or is in at most in max_position in the current research queue.
    """
    empire = fo.getEmpire()
    if empire.techResearched(tech):
        return True
    research_queue = empire.researchQueue
    return any(research_queue[i].tech == tech for i in range(0, min(max_position, len(research_queue))))


def policy_is_adopted(policy: str) -> bool:
    """
    Return if policy is currently adopted.
    """
    return fo.getEmpire().policyAdopted(policy)


def ppstring(foo):
    """
    Returns a string version of lists, dicts, sets, such that entries with special characters will be
    printed in legible string format rather than as hex escape characters, i.e.,
    ['Asimov α'] rather than ['Asimov \xce\xb1']."""

    if isinstance(foo, list):
        return "[" + ",".join(map(ppstring, foo)) + "]"
    elif isinstance(foo, dict):
        return "{" + ",".join([ppstring(k) + ":" + ppstring(v) for k, v in foo.items()]) + "}"
    elif isinstance(foo, tuple):
        return "(" + ",".join(map(ppstring, foo)) + ")"
    elif isinstance(foo, set) or isinstance(foo, frozenset):
        return "{" + ",".join(map(ppstring, foo)) + "}"
    elif isinstance(foo, str):
        return "'" + foo + "'"
    else:
        return str(foo)


class ConsoleLogHandler(Handler):
    """A log handler to send errors to the console."""

    def emit(self, record):
        """Emit a record.

        If a formatter is specified, it is used to format the record and then sent to human players."""
        try:
            human_ids = [x for x in fo.allPlayerIDs() if fo.playerIsHost(x)]
            if not human_ids:
                return
            msg = self.format(record)

            for human_id in human_ids:
                fo.sendChatMessage(human_id, msg)
        except (KeyboardInterrupt, SystemExit):
            raise
        # Hide errors from within the ConsoleLogHandler
        except:  # noqa: E722
            self.handleError(record)


# Create the log handler, format it and attach it to the root logger
console_handler = ConsoleLogHandler()

console_handler.setFormatter(
    FOLogFormatter(
        RED % ("%s : %%(filename)s:%%(funcName)s():%%(lineno)d  - %%(message)s" % fo.userString("AI_ERROR_MSG"))
    )
)

console_handler.setLevel(ERROR)

getLogger().addHandler(console_handler)


def remove_tags(message):
    """Remove tags described in Font.h from message."""
    expr = r"</?(i|u|(rgba ([0-1]\.)?\d+ ([0-1]\.)?\d+ ([0-1]\.)?\d+ ([0-1]\.)?\d+)|rgba|left|center|right|pre)>"
    return re.sub(expr, "", message)


def chat_human(message, send_to_logs=True):
    """
    Send chat message to human and print it to log.
    Log message cleared form tags.
    """
    human_id = [x for x in fo.allPlayerIDs() if fo.playerIsHost(x)][0]
    message = str(message)
    fo.sendChatMessage(human_id, message)
    if send_to_logs:
        debug("Chat Message to human: %s", remove_tags(message))


@cache_for_current_turn
def get_partial_visibility_turn(obj_id: int) -> int:
    """Return the last turn an object had at least partial visibility.

    :return: Last turn an object had at least partial visibility, -9999 if never
    """
    visibility_turns_map = fo.getUniverse().getVisibilityTurnsMap(obj_id, fo.empireID())
    return visibility_turns_map.get(fo.visibility.partial, -9999)


class ReadOnlyDict(Mapping):
    """A dict that offers only read access.

    Note that if the values of the ReadOnlyDict are mutable,
    then those objects may actually be changed.

    It is strongly advised to store only immutable objects.
    A slight protection is offered by checking for hashability of the values.

     Example usage:
     my_dict = ReadOnlyDict({1:2, 3:4})
     print my_dict[1]
     for k in my_dict:
         print my_dict.get(k, -1)
     for k in my_dict.keys():
         print my_dict[k]
     for k, v in my_dict.iteritems():
         print k, v
     my_dict[5] = 4  # throws TypeError
     del my_dict[1]  # throws TypeError

     Implementation note:

    The checks that values are hashable is the main difference from the built-in types.MappingProxyType.
    MappingProxyType has slightly different signature and cannot be inherited.
    """

    def __init__(self, *args, **kwargs):
        self._data = dict(*args, **kwargs)
        for k, v in self._data.items():
            try:
                hash(v)
            except TypeError:
                error("Tried to store a non-hashable value in ReadOnlyDict")
                raise

    def __getitem__(self, item):
        return self._data[item]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)


def dump_universe():
    """Dump the universe but not more than once per turn."""
    cur_turn = fo.currentTurn()

    if not hasattr(dump_universe, "last_dump") or dump_universe.last_dump < cur_turn:
        dump_universe.last_dump = cur_turn
        fo.getUniverse().dump()  # goes to debug logger


class LogLevelSwitcher:
    """A context manager class which controls the log level within its scope.

    Example usage:
    logging.getLogger().setLevel(logging.INFO)

    debug("Some message")  # not printed because of log level
    with LogLevelSwitcher(logging.DEBUG):
        debug("foo")  # printed because we set to DEBUG level

    debug("baz")  # not printed, we are back to INFO level
    """

    def __init__(self, log_level):
        self.target_log_level = log_level
        self.old_log_level = 0

    def __enter__(self):
        self.old_log_level = getLogger().level
        getLogger().setLevel(self.target_log_level)

    def __exit__(self, exc_type, exc_val, exc_tb):
        getLogger().setLevel(self.old_log_level)


def with_log_level(log_level):
    """A decorator to set a specific logging level for the function call.

    This decorator is useful to selectively activate debugging for a specific function
    while the rest of the code base only logs at a higher level.

    If functions are called within the decorated function, then those will be
    executed with the same logging level. However, if those functions use this
    decorator as well, then that logging level will be respected for its scope.

    Example usage:
    @log_with_specific_log_level(logging.DEBUG)
    def foo():
        debug("debug stuff")
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with LogLevelSwitcher(log_level):
                return func(*args, **kwargs)

        return wrapper

    return decorator


def assertion_fails(cond: bool, msg: str = "") -> bool:
    """
    Check if condition fails and if so, log a traceback but raise no Exception.

    Return False is condition True and True if condition is False.

    This is a useful functions for generic sanity checks and may be used to
    replace manual error logging with more context provided by the traceback.
    """
    if cond:
        return False

    if msg:
        header = "AI raised a warning: %s. See more detailed description in logs." % msg
    else:
        header = "AI raised a warning. See more detailed description in logs."

    warning("\n===")
    error(header)
    stack = traceback.extract_stack()[:-1]  # do not log this function
    warning("Stack trace (most recent call last): %s", "".join(traceback.format_list(stack)))
    frame = inspect.currentframe().f_back
    local_vars = pprint.pformat(frame.f_locals)
    warning(f"Locals inside the {frame.f_code.co_name}\n{local_vars}")
    warning("===\n")
    return True


@cache_for_session
def get_species_tag_grade(species_name: Optional[SpeciesName], tag_type: AIDependencies.Tags) -> str:
    """Determine grade string ("NO", "BAD", "GOOD", etc.), if any, for given tag and species."""
    if not species_name:
        return ""
    species = fo.getSpecies(species_name)
    if assertion_fails(species is not None):
        return ""

    return get_ai_tag_grade(species.tags, tag_type)


@cache_for_session
def get_species_stealth(species_name: Optional[SpeciesName]) -> float:
    grade = get_species_tag_grade(species_name, AIDependencies.Tags.STEALTH)
    return AIDependencies.STEALTH_STRENGTHS_BY_SPECIES_TAG.get(grade, 0.0)


@cache_for_session
def get_species_attack_troops(species_name: Optional[SpeciesName]) -> float:
    grade = get_species_tag_grade(species_name, AIDependencies.Tags.ATTACKTROOPS)
    return AIDependencies.SPECIES_TROOP_MODIFIER.get(grade, 1.0)


@cache_for_session
def get_species_fuel(species_name: Optional[SpeciesName]) -> float:
    grade = get_species_tag_grade(species_name, AIDependencies.Tags.FUEL)
    return AIDependencies.SPECIES_FUEL_MODIFIER.get(grade, 0.0)


@cache_for_session
def get_species_ship_shields(species_name: Optional[SpeciesName]) -> float:
    grade = get_species_tag_grade(species_name, AIDependencies.Tags.SHIP_SHIELDS)
    return AIDependencies.SPECIES_SHIP_SHIELD_MODIFIER.get(grade, 0.0)


@cache_for_session
def get_species_stability(species_name: Optional[SpeciesName]) -> float:
    grade = get_species_tag_grade(species_name, AIDependencies.Tags.STABILITY)
    return AIDependencies.SPECIES_STABILITY_MODIFIER.get(grade, 0.0)


@cache_for_session
def get_species_supply(species_name: Optional[SpeciesName]) -> int:
    grade = get_species_tag_grade(species_name, AIDependencies.Tags.SUPPLY)
    return int(AIDependencies.SPECIES_SUPPLY_MODIFIER.get(grade, 1))


@cache_for_session
def get_species_population(species_name: Optional[SpeciesName]) -> float:
    grade = get_species_tag_grade(species_name, AIDependencies.Tags.POPULATION)
    return AIDependencies.SPECIES_POPULATION_MODIFIER.get(grade, 1.0)


@cache_for_session
def get_species_influence(species_name: Optional[SpeciesName]) -> float:
    grade = get_species_tag_grade(species_name, AIDependencies.Tags.INFLUENCE)
    return AIDependencies.SPECIES_INFLUENCE_MODIFIER.get(grade, 1.0)


@cache_for_session
def get_species_research(species_name: Optional[SpeciesName]) -> float:
    grade = get_species_tag_grade(species_name, AIDependencies.Tags.RESEARCH)
    return AIDependencies.SPECIES_RESEARCH_MODIFIER.get(grade, 1.0)


@cache_for_session
def get_species_industry(species_name: Optional[SpeciesName]) -> float:
    grade = get_species_tag_grade(species_name, AIDependencies.Tags.INDUSTRY)
    return AIDependencies.SPECIES_INDUSTRY_MODIFIER.get(grade, 1.0)


@cache_for_session
def get_ship_part(part_name: str):
    """Return the shipPart object (fo.getShipPart(part_name)) of the given part_name.

    As the function in late game may be called some thousand times, the results are cached.
    """
    if not part_name:
        return None

    part_type = fo.getShipPart(part_name)
    if not part_type:
        warning("Could not find part %s" % part_name)

    return part_type


def get_named_int(name: str) -> int:
    """
    Returns a NamedReal from FOCS.
    If the value does not exist, reports an error and returns 1.
    Note that we do not raise and exception so that the AI can continue, as good as it can, with outdated information.
    This is also why we return 1, returning 0 could cause followup errors if the value is used as divisor.
    """
    if fo.namedIntDefined(name):
        return fo.getNamedInt(name)
    error(f"Requested integer {name} does not exist!")
    return 1


def get_named_real(name: str) -> float:
    """
    Returns a NamedReal from FOCS.
    If the value does not exist, reports an error and returns 1.0.
    Note that we do not raise and exception so that the AI can continue, as good as it can, with outdated information.
    This is also why we return 1, returning 0 could cause followup errors if the value is used as divisor.
    """
    if fo.namedRealDefined(name):
        return fo.getNamedReal(name)
    error(f"Requested integer {name} does not exist!")
    return 1.0


def get_game_rule_int(name: str, default: int) -> int:
    """
    Returns an integer value for a game rule.
    If the current game does not include the rule, the default value is returned.
    Note that unlike named values, which should always exists unless someone changes the scripting without adapting
    the AI, game rules are set when a game is started. When loading old game with a newer version, the scripting
    will use the default values of game rules not found in the save file.
    """
    rules = fo.getGameRules()
    if rules.ruleExistsWithType(name, fo.ruleType.int):
        return rules.getInt(name)
    return default


def get_game_rule_real(name: str, default: int) -> float:
    """
    Returns a real value for a game rule.
    If the current game does not include the rule, the default value is returned. See also get_game_rule_int.
    """
    rules = fo.getGameRules()
    if rules.ruleExistsWithType(name, fo.ruleType.double):
        return rules.getDouble(name)
    return default



from focs._species import *
from species.species_macros.detection import GOOD_DETECTION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import NARROW_EP, OCEAN_NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.general import SLOW_COLONIZATION
from species.species_macros.happiness import BAD_HAPPINESS
from species.species_macros.industry import BAD_INDUSTRY
from species.species_macros.influence import BAD_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import BAD_POPULATION
from species.species_macros.research import GOOD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import BAD_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import GREAT_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS, BAD_OFFENSE_TROOPS

Species(
    name="SP_HAPPY",
    description="SP_HAPPY_DESC",
    gameplay_description="SP_HAPPY_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ROBOTIC",
        "TELEPATHIC",
        "GOOD_RESEARCH",
        "BAD_INDUSTRY",
        "BAD_INFLUENCE",
        "BAD_POPULATION",
        "BAD_HAPPINESS",
        "GREAT_SUPPLY",
        "GOOD_DETECTION",
        "BAD_STEALTH",
        "BAD_OFFENSE_TROOPS",
        "PEDIA_ROBOTIC_SPECIES_CLASS",
        "PEDIA_TELEPATHIC_TITLE",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "MIMETIC_ALLOY_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "MONOPOLE_SPECIAL",
        "GAIA_SPECIAL",
        "PANOPTICON_SPECIAL",
        "RESONANT_MOON_SPECIAL",
        "PLC_ARTISAN_WORKSHOPS",
        "PLC_DIVERSITY",
    ],
    dislikes=["BLD_PLANET_CLOAK", "BLD_SPATIAL_DISTORT_GEN", "PLC_ISOLATION", "PLC_NO_SUPPLY", "PLC_RACIAL_PURITY"],
    effectsgroups=[
        *BAD_INDUSTRY,
        *GOOD_RESEARCH,
        *BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *BAD_POPULATION,
        *BAD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_HAPPY"),
        *GREAT_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *BAD_OFFENSE_TROOPS,
        *BAD_STEALTH,
        *GOOD_DETECTION,
        SLOW_COLONIZATION,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        NARROW_EP,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=OCEAN_NARROW_EP,
    graphic="icons/species/ichthyoid-06.png",
)



from logging import info

from common.configure_logging import redirect_logging_to_freeorion_logger

# Logging is redirected before other imports so that import errors appear in log files.
redirect_logging_to_freeorion_logger()


class ChatHistoryProvider:
    def __init__(self):
        """
        Initializes ChatProvider. Doesn't accept arguments.
        """
        info("Chat initialized")

    def load_history(self):
        """
        Loads chat history from external sources.

        Returns list of tuples:
            (unixtime timestamp, player name, text, text color of type tuple with 4 elements)
        """
        info("Loading history...")
        # c = (255, 128, 128, 255)
        # e = (123456789012, "P1", "Test1", c)
        # return [e]
        return []

    def put_history_entity(self, timestamp, player_name, text, text_color):
        """
        Put chat into external storage.

        Return True if successfully stored. False otherwise.

        Arguments:
        timestamp -- unixtime, number of seconds from 1970-01-01 00:00:00 UTC
        player_name -- player name
        text -- chat text
        text_color -- tuple wit 4 elements
        """
        info(f"Chat {player_name}: {text} {text_color}")
        return True



from __future__ import annotations

import freeOrionAIInterface as fo
from collections.abc import Iterable
from copy import copy
from logging import debug, error
from typing import Callable, NamedTuple

import PlanetUtilsAI
from AIDependencies import Tags
from aistate_interface import get_aistate
from common.fo_typing import PlanetId, SpeciesName
from DiplomaticCorp import get_diplomatic_status
from EnumsAI import FocusType, PriorityType
from freeorion_tools import (
    assertion_fails,
    get_named_int,
    get_named_real,
    get_species_influence,
)
from freeorion_tools.caching import cache_for_current_turn
from freeorion_tools.statistics import stats
from ResearchAI import research_now
from turn_state import get_empire_planets_by_species, luxury_resources

economic_category = "ECONOMIC_CATEGORY"
industrialism = "PLC_INDUSTRIALISM"
technocracy = "PLC_TECHNOCRACY"
centralization = "PLC_CENTRALIZATION"  # currently unused
bureaucracy = "PLC_BUREAUCRACY"
capital_markets = "PLC_CAPITAL_MARKETS"
black_markets = "PLC_BLACK_MARKET"
traffic_control = "PLC_TRAFFIC_CONTROL"
infra1 = "PLC_PLANETARY_INFRA"
infra2 = "PLC_SYSTEM_INFRA"
infra3 = "PLC_INTERSTELLAR_INFRA"

military_category = "MILITARY_CATEGORY"
allied_repair = "PLC_ALLIED_REPAIR"
charge = "PLC_CHARGE"
scanning = "PLC_CONTINUOUS_SCANNING"
simplicity = "PLC_DESIGN_SIMPLICITY"
engineering = "PLC_ENGINEERING"
exploration = "PLC_EXPLORATION"
flanking = "PLC_FLANKING"
recruitment = "PLC_MARINE_RECRUITMENT"
# martial law and terror suppression currently ignored

social_category = "SOCIAL_CATEGORY"
propaganda = "PLC_PROPAGANDA"
algo_research = "PLC_ALGORITHMIC_RESEARCH"
liberty = "PLC_LIBERTY"
diversity = "PLC_DIVERSITY"
artisans = "PLC_ARTISAN_WORKSHOPS"
population = "PLC_POPULATION"
racial_purity = "PLC_RACIAL_PURITY"
conformance = "PLC_CONFORMANCE"
conformance_exclusions = {liberty, diversity, artisans}

initial_influence_priority = 18.0


class _EmpireOutput:
    def __init__(self, has_liberty: bool):
        self.industry = 0.0
        self.research = 0.0
        self.influence = 0.0
        self.population_stability = 0.0
        self.stability_scaling = 0.5
        self.has_liberty = has_liberty

    def __str__(self):
        return "pp=%.1f, rp=%.1f, ip=%.1f, population_stabililty=%d" % (
            self.industry,
            self.research,
            self.influence,
            self.population_stability,
        )

    def difference(self, other: _EmpireOutput) -> float:
        """Return how much this is better (>0) or worse (<0) than other."""
        aistate = get_aistate()
        delta_pp = (self.industry - other.industry) * aistate.get_priority(PriorityType.RESOURCE_PRODUCTION)
        delta_rp = (self.research - other.research) * aistate.get_priority(PriorityType.RESOURCE_RESEARCH)
        delta_ip = (self.influence - other.influence) * aistate.get_priority(PriorityType.RESOURCE_INFLUENCE)
        delta_stability = self.population_stability - other.population_stability
        result = delta_pp + delta_rp + delta_ip + delta_stability
        debug(f"difference: {delta_pp} + {delta_rp} + {delta_ip} + {delta_stability} => {result}")
        return result

    def is_better_than(self, other: _EmpireOutput, cost_for_other: float) -> bool:
        """
        Return True if this is better than changing to other at given IP cost.
        I.e. if is_better_than returns False, changing to other is considered worth the cost.
        """
        # delta is per round, adoption cost is a one-time cost
        cost = cost_for_other * get_aistate().get_priority(PriorityType.RESOURCE_INFLUENCE) / 10
        return self.difference(other) + cost > 0

    def add_planet(self, planet: fo.planet) -> None:
        """Add output of the given planet to this object."""
        self.industry += planet.currentMeterValue(fo.meterType.targetIndustry)
        self.research += planet.currentMeterValue(fo.meterType.targetResearch)
        self.influence += planet.currentMeterValue(fo.meterType.targetInfluence)
        current_population = planet.currentMeterValue(fo.meterType.population)
        target_population = planet.currentMeterValue(fo.meterType.targetPopulation)
        # current is what counts now, but look a little ahead
        population = (3 * current_population + target_population) / 4
        # no bonuses above 20
        stability = min(planet.currentMeterValue(fo.meterType.targetHappiness), 20)
        if stability < -1:
            # increase both for the risk of losing the colony
            population += 1
            stability -= 1
        elif stability >= 10:
            # 10 give a lot of bonuses, increases above 10 are less important
            stability = 11 + (stability - 10) / 2
        if self.has_liberty and planet.focus == FocusType.FOCUS_RESEARCH:
            self.research += PlanetUtilsAI.adjust_liberty(planet, population)
        self.population_stability += population * stability * self.stability_scaling


class _Alternative(NamedTuple):
    output: _EmpireOutput
    costs: float
    adopted: set


class PolicyManager:
    """Policy Manager for one round"""

    def __init__(self, status_only: bool = False):
        self._empire = fo.getEmpire()
        self._universe = fo.getUniverse()
        self._aistate = get_aistate()
        # resourceAvailable includes this turns production, but that is wrong for influence
        self._ip = self._empire.resourceAvailable(fo.resourceType.influence) - self._get_infl_prod()
        self._adopted = set(self._empire.adoptedPolicies)
        # When we continue a game in which we just adopted a policy, game state shows the policy as adopted,
        # but IP still unspent. Correct it here, then calculate anew whether we want to adopt it.
        if not status_only:
            for entry in self._empire.turnsPoliciesAdopted:
                if entry.data() == fo.currentTurn():
                    debug(f"reverting saved adopt {entry.key()}")
                    fo.issueDeadoptPolicyOrder(entry.key())
                    self._adopted.remove(entry.key())
        self._originally_adopted = copy(self._adopted)
        empire_owned_planet_ids = PlanetUtilsAI.get_owned_planets_by_empire()
        self._populated_planet_ids = PlanetUtilsAI.get_populated_planet_ids(empire_owned_planet_ids)
        self._num_populated = len(self._populated_planet_ids)
        self._num_outposts = len(empire_owned_planet_ids) - self._num_populated
        self._wanted_ip = 0.0
        self._adoptable = self._get_adoptable()
        self._available = set(self._empire.availablePolicies)
        self._rating_functions = {
            propaganda: self._rate_propaganda,
            algo_research: self._rate_algo_research,
            diversity: self._rate_diversity,
            artisans: self._rate_artisans,
            population: self._rate_population,
            liberty: lambda: 10,  # hack, only used by _try_without_conformance
            conformance: lambda: 999,  # not an actual rating, but _deadopt_one_of fails without it
            # military policies are mostly chosen by opinion, plus some rule-of-thumb values
            allied_repair: lambda: self._rate_opinion(charge),  # no effect, unless we have allies...
            charge: lambda: self._rate_opinion(charge),  # increases attack, reduces shields
            scanning: lambda: 10 + self._rate_opinion(scanning),  # May help us detect stealthy attackers
            simplicity: lambda: 20 + self._rate_opinion(simplicity),  # Makes simple ships cheaper
            engineering: self._rate_engineering_corps,
            exploration: lambda: 10 + self._rate_opinion(exploration),  # may give a little research, speeds up scouts
            flanking: lambda: 5 + self._rate_opinion(flanking),  # May give small a bonus in battle
            recruitment: lambda: 15 + self._rate_opinion(recruitment),  # cheaper troop ships
            # economic policies
            bureaucracy: self._rate_bureaucracy,
            capital_markets: self._rate_capital_markets,
            black_markets: self._rate_black_markets,
            traffic_control: lambda: 10 + self._rate_opinion(traffic_control),  # slightly faster ships
        }

    # Policies that may use IP "reserved" by wanted_ip.
    # The first two get IP specially reserved for them in the constructor.
    # If we are low on IP, adopting something that helps as gain IP should be worth it.
    # Technocracy and industrialism are important overall.
    _prioritised_policies = (propaganda, artisans, technocracy, industrialism)

    # This value is increased by ResourceAI via report_focus_change(). Generate orders copies the value into
    # a member variable and reset this value so that we can count anew next turn.
    _num_focus_changes = 0

    @staticmethod
    def report_focus_change() -> None:
        """
        Function for ResourceAI to report number of focus changes done this turn.
        Information affects whether we adapt or deadopt bureaucracy.
        """
        PolicyManager._num_focus_changes += 1

    def generate_orders(self) -> None:
        """The main task of the class, called once every turn by FreeOrionAI."""
        debug(
            "Start turn %d IP: %.2f + %.2f, adopted: %s, adoptable: %s",
            fo.currentTurn(),
            self._ip,
            self._get_infl_prod(),
            self._adopted,
            self._adoptable,
        )
        debug(f"Empty slots: {[str(s) for s in self._empire.emptyPolicySlots]}")
        self._process_social()
        self._process_military()
        self._process_economic()
        new_production = self._get_infl_prod(True)
        debug("End of turn IP: %.2f (wanted %.2f) + %.2f", self._ip, self._wanted_ip, new_production)
        self._determine_influence_priority(new_production)
        PolicyManager._num_focus_changes = 0

    def _process_social(self) -> None:
        """
        Process social policies.
        Propaganda is activated in turn 1 and usually replaced by liberty in 7.
        Since liberty and conformance affect so many other things, we check them first, then simply rate all others and
        try to adopt the best possible ones for the slots we have.
        """
        self._conformance_or_liberty()
        options = {propaganda, algo_research, diversity, artisans, population}
        if conformance in self._adopted:
            options -= conformance_exclusions
        self._process_policy_options(social_category, options)

    def _process_military(self) -> None:
        """Process military policies."""
        options = (allied_repair, charge, scanning, simplicity, engineering, exploration, flanking, recruitment)
        self._process_policy_options(military_category, options)

    def _process_economic(self) -> None:
        """Process economic policies."""
        # We need the extra slots, before adopting others economic policies.
        # Starting with centralization may make sense, but it's too complicated to program for the little gain.
        if infra1 in self._adopted:
            self._techno_or_industry()
            self._process_infrastructure23()
            if technocracy not in self._empire.availablePolicies and self._can_adopt(bureaucracy):
                # we have to adopt bureaucracy at least once at the beginning to unlock technocracy and IRAs
                self._adopt(bureaucracy)
            else:
                options = (bureaucracy, capital_markets, black_markets, traffic_control)
                self._process_policy_options(economic_category, options)
        elif infra1 in self._adoptable:
            self._adopt(infra1)

    def _process_policy_options(self, category: str, options: Iterable[str]) -> None:  # noqa: C901
        """
        Rate all given policies and deadopt policies with a negative rating.
        Then evaluate if any of the other is worth being adopted, possible even to replace another one.
        If we find a policy should be adopted, but lack the IP, increase _wanted_ip.
        Given policies must all have the same category.
        """
        adopted = []
        may_adopt = []
        for policy in options:
            if not self._potentially_available(policy):
                continue
            rating = self._rate_policy(policy)
            cost = fo.getPolicy(policy).adoptionCost()
            debug(f"_process_policy_options {policy} adopted={policy in self._adopted} rating={rating} cost={cost}")
            if policy in self._adopted:
                if rating < 0:
                    self._deadopt(policy)
                else:
                    adopted.append((rating, policy))
            elif rating > 6:
                may_adopt.append((rating, policy, cost))
        if may_adopt:
            adopted.sort()
            may_adopt.sort()
            # Policies do not change too often, one per turn should do
            new_rating, new_policy, new_cost = may_adopt.pop()
            rated_cost = self._rate_costs(new_cost)
            debug(
                f"_process_policy_options({category}): best new={new_policy}, rating={new_rating}, cost={new_cost}"
                f", rated_cost={rated_cost}, slots={self._empire.emptyPolicySlots[category]}"
            )
            if self._empire.emptyPolicySlots[category]:
                if new_rating > rated_cost:
                    self._we_want(new_policy, new_cost)
            elif adopted:
                old_rating, old_policy = adopted.pop(0)
                debug(f"_process_policy_options({category}): lowest adopted={old_policy}, rating={old_rating}")
                if new_rating - old_rating > rated_cost:
                    self._we_want(new_policy, new_cost, old_policy)

    def _we_want(self, name: str, cost: float, replace: str | None = None) -> None:
        if name not in self._empire.availablePolicies:
            self._try_unlock(name)
            return
        if self._can_adopt(name, replace):
            if replace:
                self._deadopt(replace)
            self._adopt(name)
        else:
            # We seem to have insufficient IP. Add cost to wanted_ip, to raise the influence priority
            self._wanted_ip += cost
            debug(f"Adopting {name} failed due to insufficient IP, adding {cost} to wanted_ip")

    def _unlocked_by(self, name: str) -> str | Callable | None:
        """
        Returns another policy that unlocks the given policy, a function to unlock it, or None.
        Returned functions can be called with False to check whether it could unlock the given policy.
        """
        if name == artisans:
            # Some day we should get that information from the server, but then, if it changes,
            # we need to adapt the AI anyway.
            return diversity
        if name == population:
            return self._unlock_population
        return None

    def _try_unlock(self, name: str) -> None:
        """Unlock the given policy, if possible. Note that unlocking always takes at least one turn."""
        unlocker = self._unlocked_by(name)
        if isinstance(unlocker, str):
            debug(f"Trying to adopt {unlocker} to unlock {name}")
            self._we_want(unlocker, fo.getPolicy(unlocker).adoptionCost())
        else:
            unlocker(True)

    def _unlock_population(self, do_unlock: bool) -> bool:
        """
        Unlock population by putting GRO_NANOTECH_MED and its prerequisites at the top of the research queue.
        Until Adaptive Automation has been researched, we do not want to use research points for unlocking
        population. Also, AA requires Nanotech Production, which is the most expensive prerequisite
        of Nanotech Medicine. So if AA has not been researched, this function only returns false.
        If do_unlock is false, the function only returns whether we could unlock it, i.e. AA is available.
        """
        # TBD: create Enum for technologies
        could_do = self._empire.techResearched("PRO_ADAPTIVE_AUTOMATION")
        if could_do and do_unlock:
            unlocker = "GRO_NANOTECH_MED"
            debug(f"Prioritising research for {unlocker} to unlock population")
            research_now(unlocker, True)
        return could_do

    @cache_for_current_turn
    def _rate_policy(self, name: str) -> float:
        """Gives a rough value of a policy."""
        rating_function = self._rating_functions.get(name)
        if assertion_fails(rating_function, f"_rate_policy({name}) not yet supported"):
            return 0.0
        return rating_function()

    def _rate_costs(self, costs: float) -> float:
        """
        How do we rate the costs in comparison to a policies rating?
        Generally costs / 3, but less if we have plenty of IP.
        """
        # avoid division by zero, with unreserved_ip < 1 the policy won't be adopted anyway
        unreserved_ip = max(1, self._ip - self._wanted_ip)
        # For big empires, costs of some policies go over 100 and their rating doesn't grow that much, but they may
        # also have lots of IP and the policy may still be useful. Note that costs should be less than unreserved_ip,
        # so if cost is e.g. 1/3 of unreserved_ip, this limits the rating to 0.5 * priority.
        return min(costs / 3, 1.5 * self._get_influence_priority() * costs / unreserved_ip)

    @staticmethod
    def _rate_opinion(name: str) -> float:
        """Return a rating value for the empire's planet opinion on a policy"""
        opinion = PlanetUtilsAI.get_planet_opinion(name)
        return 1.5 * (len(opinion.likes) - len(opinion.dislikes) * PlanetUtilsAI.dislike_factor())

    def _rate_propaganda(self) -> float:
        """Rate propaganda broadcasts"""
        ip_gain = get_named_real("PROPAGANDA_INFLUENCE_FLAT")
        ip_enemy = get_named_real("PROPAGANDA_INFLUENCE_ENEMY_FLAT")
        # Is there an easy way to determine how many enemies we see? For simplicity assume we see half of them.
        # Note that the enemy effect may be our gain (if enemy also uses propaganda), or the other player's loss.
        num_enemies = sum(
            0.5 for e in fo.allEmpireIDs() if e != fo.empireID() and get_diplomatic_status(e) == fo.diplomaticStatus.war
        )
        rating = self._get_influence_priority() * (ip_gain - num_enemies * ip_enemy) + self._rate_opinion(propaganda)
        debug(f"_rate_propaganda: rating={rating}")
        return rating

    def _rate_algo_research(self) -> float:
        """Rate algorithmic research"""
        min_stability = get_named_real("LRN_ALGO_RESEARCH_MIN_STABILITY")
        scaling = get_named_real("LRN_ALGO_RESEARCH_TARGET_RESEARCH_PERCONSTRUCTION")
        research_priority = self._aistate.get_priority(PriorityType.RESOURCE_RESEARCH)
        gain = sum(
            scaling
            * research_priority
            * min(
                planet.currentMeterValue(fo.meterType.population),
                planet.currentMeterValue(fo.meterType.construction),
            )
            for planet in PlanetUtilsAI.get_empire_populated_planets()
            if planet.focus == FocusType.FOCUS_RESEARCH
            and planet.currentMeterValue(fo.meterType.targetHappiness) >= min_stability
        )
        rating = gain + self._rate_opinion(algo_research)
        debug(f"_rate_algo_research: rating={rating}")
        return rating

    def _conformance_or_liberty(self) -> _EmpireOutput:
        """
        Try policies conformance and liberty.
        Liberty is available very early, but to help setting up basic policies we keep propaganda at least
        until we have infra1. Getting a second slot before that should be nearly impossible.
        """
        if infra1 in self._adopted:
            debug("Evaluating conformance/liberty, first without change:")
            current_output = self._calculate_empire_output()
            if conformance in self._adopted:
                # this may also try liberty
                self._try_without_conformance(current_output)
            else:
                current_output, conformance_cost = self._try_conformance(current_output)
                self._process_liberty(current_output, conformance_cost)

    def _try_without_conformance(self, current_output: _EmpireOutput) -> None:
        """
        Consider de-adopting conformance.
        Conformance always improves stability (even when disliked). Theoretically this could be negative when
        using black markets and/or necessity. So far we ignore this and only try de-adopting it when we can
        adopt one or more of its exclusions instead.
        """
        if not any(self._can_adopt(p, conformance) for p in conformance_exclusions):
            return
        self._deadopt(conformance)
        # For this we use a fixed, low rating for liberty since later in the game (it must be later, since we have
        # conformance) it is usually not so great. Even more so when the empire was using conformance.
        adopt_options = sorted([(self._rate_policy(p), p) for p in conformance_exclusions], reverse=True)
        best = _Alternative(current_output, 0.0, set())
        costs = 0.0
        adopted = set()
        # Note that this function does not unlock artisans. Perhaps it should?
        # _process_policy_options cannot deadopt conformance to do it.
        for rating, policy in adopt_options:
            if rating > 0.0 and self._can_adopt(policy):
                self._adopt(policy)
                costs += fo.getPolicy(policy).adoptionCost()
                adopted.add(policy)
                new_output = self._calculate_empire_output()
                if new_output.is_better_than(best.output, costs - best.costs):
                    best = _Alternative(new_output, costs, adopted)
        for policy in adopted - best.adopted:
            # best wasn't the last one, so undo all further adoptions
            self._deadopt(policy)
        if not best.adopted:
            # No alternative was better than current_output with conformance.
            self._adopt(conformance)

    def _try_conformance(self, current_output: _EmpireOutput) -> tuple[_EmpireOutput, float]:
        """
        Try adopting conformance, may replace another social policies.
        Conformance is rather good, but has several exclusions. So this may have to de-adopt all exclusions just
        to try adoption conformance.
        """
        exclusions = conformance_exclusions & self._adopted
        other_could_replace = {propaganda, algo_research, population} & self._adopted
        adoption_cost = fo.getPolicy(conformance).adoptionCost()
        if self._can_adopt(conformance):
            self._adopt(conformance)
            with_conformance = self._calculate_empire_output()
            if current_output.is_better_than(with_conformance, adoption_cost):
                self._deadopt(conformance)
                return current_output, 0
            return with_conformance, adoption_cost
        elif self._can_adopt(conformance, exclusions | other_could_replace):
            if exclusions:
                for policy in exclusions:
                    self._deadopt(policy)
            else:
                self._deadopt_one_of(other_could_replace)
            self._adopt(conformance)
            with_conformance = self._calculate_empire_output()
            if current_output.is_better_than(with_conformance, adoption_cost):
                self._deadopt(conformance)
                if exclusions:
                    for policy in exclusions:
                        self._adopt(policy)
                else:
                    self._readopt_selected_one()
                return current_output, 0
            return with_conformance, adoption_cost
        # cannot adopt it at all
        return current_output, 0

    def _process_liberty(self, current_output: _EmpireOutput, conformance_cost: float) -> None:
        """
        Adopt or deadopt liberty, may replace another social policy.
        """
        could_replace = {propaganda, algo_research, diversity, artisans} & self._adopted
        if conformance in self._adopted:
            could_replace = {conformance}  # cannot adopt them together
        if liberty in self._adopted or self._can_adopt(liberty, could_replace):
            adoption_cost = 0.0 if liberty in self._originally_adopted else fo.getPolicy(liberty).adoptionCost()
            # liberty will generally make species less happy, but generates research
            if liberty in self._adopted:
                self._deadopt(liberty)
                without_liberty = self._calculate_empire_output()
                if current_output.is_better_than(without_liberty, 0.0):
                    self._adopt(liberty)
            elif self._can_adopt(liberty):
                self._adopt(liberty)
                with_liberty = self._calculate_empire_output()
                if current_output.is_better_than(with_liberty, adoption_cost):
                    self._deadopt(liberty)
            else:
                # Can only adopt it by replacing something. Note that we ignore the other replaced policies rating
                # here, since its effect is also evaluated by _calculate_empire_output.
                self._deadopt_one_of(could_replace)
                self._adopt(liberty)
                with_liberty = self._calculate_empire_output()
                # Current cost is only relevant when we adopted conformance in _try_conformance. Then we must compare
                # current_output with conformance to with_liberty at the cost difference of cost.
                if current_output.is_better_than(with_liberty, adoption_cost - conformance_cost):
                    self._deadopt(liberty)
                    self._readopt_selected_one()

    def _calculate_empire_output(self) -> _EmpireOutput:
        """Update empire meters, then calculate an estimation of the expected empire output."""
        # TBD: here we could use the stability-adapted update, that would be much better than trying
        # to rate the stability effects again production effects.
        self._universe.updateMeterEstimates(self._populated_planet_ids)
        result = _EmpireOutput(liberty in self._adopted)
        for pid in self._populated_planet_ids:
            result.add_planet(self._universe.getPlanet(pid))
        debug(f"Empire output: {result}")
        return result

    def _rate_diversity(self) -> float:
        """Rate diversity."""
        if diversity not in self._empire.availablePolicies:
            return 0.0
        # diversity affects stability, but also gives a bonus to research-focused planets and a little influence,
        diversity_value = len(get_empire_planets_by_species()) - get_named_int("PLC_DIVERSITY_THRESHOLD")
        diversity_scaling = get_named_real("PLC_DIVERSITY_SCALING")
        # Research bonus goes to research-focused planets only. With priority there are usually none.
        research_priority = self._aistate.get_priority(PriorityType.RESOURCE_RESEARCH)
        research_bonus = max(0, research_priority - 20)
        # + 4 for global influence
        global_influence_bonus = 4
        rating = self._rate_opinion(diversity) + diversity_scaling * diversity_value * (
            self._num_populated + global_influence_bonus + research_bonus
        )
        debug(
            f"_rate_diversity: rating={rating}. diversity_value={diversity_value}, "
            f"research priority={research_priority}"
        )
        return rating

    def _rate_artisans(self) -> float:
        """Rate artisan workshops."""
        # Diversity unlocks artisans, so if we have diversity, we could unlock artisans
        if diversity not in self._empire.availablePolicies:
            return 0.0
        # TBD: could use _calculate_empire_output, but we'd still have to check the species, since artistic
        # species may not be switched to influence.
        rating = 0.0
        artists = []
        for species_name, planets in get_empire_planets_by_species().items():
            species = fo.getSpecies(species_name)
            if Tags.ARTISTIC in species.tags:
                artists.append(species_name)
                for pid in planets:
                    rating += self._rate_artisan_planet(pid, species_name)
        rating += self._rate_opinion(artisans)
        debug(f"_rate_artisans: {rating}, artists: {artists}")
        return rating

    def _rate_artisan_planet(self, pid: PlanetId, species_name: SpeciesName) -> float:
        focus_bonus = get_named_real("ARTISANS_INFLUENCE_FLAT_FOCUS")
        focus_minimum = get_named_real("ARTISANS_MIN_STABILITY_FOCUS")
        species_focus_bonus = focus_bonus * get_species_influence(species_name)
        planet = self._universe.getPlanet(pid)
        stability = planet.currentMeterValue(fo.meterType.targetHappiness)
        # First check whether the planet would currently get the focus bonus.
        if planet.focus == FocusType.FOCUS_INFLUENCE:
            return 3 * species_focus_bonus if stability >= focus_minimum else 0.0

        # Planet does not have influence focus...
        # Check for the non-focus bonus. Since we would get this "for free", rate it higher
        non_focus_bonus = get_named_real("ARTISANS_INFLUENCE_FLAT_NO_FOCUS")
        non_focus_minimum = get_named_real("ARTISANS_MIN_STABILITY_NO_FOCUS")
        rating = 0.0
        if stability >= non_focus_minimum:
            rating += 4 * non_focus_bonus
        # Check whether this planet would get the focus bonus, if we'd switch it to influence.
        if PlanetUtilsAI.stability_with_focus(planet, FocusType.FOCUS_INFLUENCE) >= focus_minimum:
            rating += species_focus_bonus
        return rating

    def _rate_engineering_corps(self) -> float:
        """
        Rate Engineering Corps.
        This policy reduces ship upkeep at the cost of influence, it won't do us any good if we are short on
        influence, but otherwise may help us maintain big fleets.
        """
        ships_owned = self._empire.totalShipsOwned
        debug(f"totalShipsOwned={ships_owned}")
        influence_priority = self._get_influence_priority()
        # since influence_priority tends to vary strongly, do not return negative unless
        influence_priority_threshold = 40
        if influence_priority > 2 * influence_priority_threshold:
            return -1  # any negative value means, deadopt
        if influence_priority > influence_priority_threshold:
            return 0  # do not deadopt yet
        opinion = self._rate_opinion(engineering)
        return opinion + ships_owned / 25 * (influence_priority_threshold - influence_priority)

    def _rate_population(self) -> float:
        rating = self._rate_opinion(population)
        for pid in self._populated_planet_ids:
            planet = self._universe.getPlanet(pid)
            current_population = planet.currentMeterValue(fo.meterType.population)
            target_population = planet.currentMeterValue(fo.meterType.targetPopulation)
            ratio = min(1.0, current_population / max(target_population, 0.01))
            empty_weight = 5
            # Almost empty_weight for a newly found colony on a big planet, half weight for half full, etc.
            rating += empty_weight * (1 - ratio)
        return rating

    def _rate_bureaucracy(self) -> float:
        if bureaucracy not in self._adopted and not self._can_adopt(bureaucracy):
            return 0.0
        debug(f"Evaluating bureaucracy, first without change. Focus changes: {self._num_focus_changes}")
        current_output = self._calculate_empire_output()
        focus_change_cost = self._bureaucracy_focus_change_penalty() * self._num_focus_changes
        if bureaucracy in self._adopted:
            self._deadopt(bureaucracy)
            without_bureaucracy = self._calculate_empire_output()
            rating = current_output.difference(without_bureaucracy)
            self._adopt(bureaucracy)
        else:
            self._adopt(bureaucracy)
            with_bureaucracy = self._calculate_empire_output()
            rating = with_bureaucracy.difference(current_output)
            self._deadopt(bureaucracy)
        return rating - focus_change_cost

    @staticmethod
    def _could_be_set_to_influence(planet: fo.planet) -> bool:
        """Does the given planet has a population that could be set to influence focus?"""
        return planet.speciesName and FocusType.FOCUS_INFLUENCE in fo.getSpecies(planet.speciesName).foci

    def _rate_capital_markets(self) -> float:
        """This has a lot of effects, add a fixed value for others, check the luxury export and influence dept."""
        rating = self._rate_opinion(capital_markets) + 10
        if self._ip < 0.0:
            # when we have an influence dept, capital_markets increases the penalty
            rating += self._ip * get_named_real("CAPITAL_MARKETS_DEBT_INSTABILITY") * 10
        for special, planets in luxury_resources().items():
            likes = PlanetUtilsAI.get_planet_opinion(special).likes
            # TBD: check for supply chains? Well, there are many more places we could/should do that.
            # Also, when we do, we'd have to be careful not to change everything just because an enemy
            # disrupts our supply chain temporarily.
            if any(planet.focus == FocusType.FOCUS_INFLUENCE for planet in planets):
                rating += 1.5 * len(likes)
            elif any(self._could_be_set_to_influence(planet) for planet in planets):
                rating += 0.5 * len(likes)
        return rating

    def _rate_black_markets(self) -> float:
        """This is a tricky one since it requires planets with low stability. But it may be useful."""
        max_stability = get_named_real("PLC_BLACK_MARKET_MAX_STABILITY")
        bonus = get_named_real("SPECIAL_BLACK_MARKET_INFLUENCE_FOCUS_BONUS")
        gain = 0.0
        for special, planets in luxury_resources().items():
            likes = PlanetUtilsAI.get_planet_opinion(special).likes
            for planet in PlanetUtilsAI.get_empire_populated_planets():
                if planet.id in likes:
                    skill = get_species_influence(planet.speciesName)
                    if planet.focus == FocusType.FOCUS_INFLUENCE:
                        if planet.currentMeterValue(fo.meterType.targetHappiness) <= max_stability:
                            gain += bonus * skill
                    else:
                        if (
                            self._could_be_set_to_influence(planet)
                            and PlanetUtilsAI.stability_with_focus(planet, FocusType.FOCUS_INFLUENCE) <= max_stability
                        ):
                            gain += bonus * skill / 3.0
        return self._rate_opinion(black_markets) + gain * 4.0

    def _techno_or_industry(self) -> None:
        """Adopt technocracy or industrialism, depending on AI priorities."""
        research_prio = self._aistate.get_priority(PriorityType.RESOURCE_RESEARCH)
        production_prio = self._aistate.get_priority(PriorityType.RESOURCE_PRODUCTION)
        may_switch_to_techno = self._can_adopt(technocracy, industrialism)
        may_switch_to_industry = self._can_adopt(industrialism, technocracy)
        # do not switch too often
        if technocracy in self._adopted:
            research_prio *= 1.25
        elif industrialism in self._adopted:
            production_prio *= 1.25
        if may_switch_to_industry and research_prio < production_prio:
            self._deadopt(technocracy)
            self._adopt(industrialism)
        elif may_switch_to_techno and research_prio > production_prio:
            self._deadopt(industrialism)
            self._adopt(technocracy)
        debug(
            "Prio res./ind. = %.1f/%.1f, maySwitch = %s/%s",
            research_prio,
            production_prio,
            may_switch_to_techno,
            may_switch_to_industry,
        )

    @cache_for_current_turn
    def _bureaucracy_focus_change_penalty(self):
        """
        How much influence do we lose due to a focus change while bureaucracy is adopted?
        With bureaucracy a focus change set influence to -3, the overall effect depends on how quickly influence
        productions raises.
        """
        have_capital_markets = capital_markets in self._adopted
        have_transorganic_sentience = self._empire.techResearched("GRO_TRANSORG_SENT")
        cost = 3.0 if have_capital_markets else 4.0 if have_transorganic_sentience else 6.0
        # this is a one-time effect, so do not rate it too high
        return cost * 0.5

    def _process_infrastructure23(self) -> None:
        """Try to adopt infrastructure2 and 3."""
        if infra3 in self._adopted:
            return
        # Note that bureaucracy is not in the list, because de-adopting it will likely trigger a some focus changes,
        # which will then get penalised when we re-adopt it the next turn.
        could_replace = {industrialism, technocracy, capital_markets, black_markets, traffic_control} & self._adopted
        if infra2 not in self._adopted:
            if self._can_adopt(infra2, could_replace) and self._get_influence_priority() < 18:
                self._maybe_adopt_infrastructure(infra2, could_replace)
        else:
            if self._can_adopt(infra3, could_replace) and self._get_influence_priority() < 12:
                self._maybe_adopt_infrastructure(infra3, could_replace)

    def _maybe_adopt_infrastructure(self, policy: str, could_replace: Iterable[str]) -> None:
        """
        Adopts the specified policy, if we have a free slot.
        Without a slot, it deadopts the cheapest of could_replace, but only if we have enough IP to readopt it
        next turn with some more to spare. Note that adopting infra3 gives us an extra slot, so we may want
        to adopt two policies afterwards.
        """
        if self._can_adopt(policy):
            self._adopt(policy)
        elif could_replace:
            # deadopt the cheapest when we have enough IP to re-adopt it next turn without getting too low.
            choices = sorted([(fo.getPolicy(name).adoptionCost(), name) for name in could_replace])
            policy_cost = fo.getPolicy(policy).adoptionCost()
            if self._ip - self._wanted_ip >= policy_cost + 2 * choices[0][0]:
                self._deadopt(choices[0][1])
                self._adopt(policy)

    def _determine_influence_priority(self, new_production: float) -> None:
        """Determine and set influence priority."""
        # avoid wildly varying priorities while we adopt the basics: simply a fixed value for some turns
        if fo.currentTurn() <= 15:
            if fo.currentTurn() == 1:
                # I'd prefer 20, but that makes Abadoni switch to Influence in turn 3.
                # TBD work on ResourceAI again...
                self._set_priority(initial_influence_priority, True)
            return
        # How much IP would we have available if we keep the current production for 3 turns?
        forecast = self._ip - self._wanted_ip + 3 * new_production
        threshold = 20
        if forecast < threshold:
            # Basic value: the lower the forecast, the more urgent we need influence
            priority = 20 + threshold - forecast
        else:
            # For higher values, slowly decrease priority
            priority = 20 * (threshold / forecast) ** 0.6
        # The more planets we have, the more influence we need in general, outpost have only a minor effect
        priority *= (3 + self._num_populated + self._num_outposts / 10) / 10
        if infra3 in self._adopted:
            priority /= 1.5  # we've done the big step, no need to amass so much anymore
        self._set_priority(priority, False)

    def _set_priority(self, calculated_priority: float, ignore_old: bool) -> None:
        """Set and log influence priority."""
        if ignore_old:
            new_priority = calculated_priority
        else:
            old_priority = self._get_influence_priority()
            # to further smoothen the values, only use 1/3 of the calculated value
            new_priority = (2 * old_priority + calculated_priority) / 3
        debug("Setting influence priority to %.1f, turn %d", new_priority, fo.currentTurn())
        self._aistate.set_priority(PriorityType.RESOURCE_INFLUENCE, new_priority)

    def _get_infl_prod(self, update=False) -> float:
        """Get / Update IP production."""
        if update:
            self._universe.updateMeterEstimates(self._populated_planet_ids)
            fo.updateResourcePools()
        return self._empire.resourceProduction(fo.resourceType.influence)

    def _potentially_available(self, name: str) -> bool:
        """Determine whether we have a way to adopt given policy."""
        if name in self._available:
            return True
        unlocker = self._unlocked_by(name)
        if unlocker is None:
            return False
        if isinstance(unlocker, str):
            return unlocker in self._available
        else:
            return unlocker(False)

    def _can_adopt(self, name: str, replace: str | set[str] | None = None) -> bool:
        """
        Can we adopt named policy, possibly by replacing (one of) replace?
        Note that when replace_other is set, this function currently assumes that with
        replace_other removed, there are no other exclusions for policy(name).
        If this won't do, the function needs to temporarily deadopt replace_other or
        policyPrereqsAndExclusionsOK must be extended so that replace_other can be passed to it.
        """
        if name in self._adopted:
            return False
        policy = fo.getPolicy(name)
        wanted = 0.0 if name in self._prioritised_policies else self._wanted_ip
        ret = name in self._empire.availablePolicies and (
            policy.adoptionCost() <= (self._ip - wanted) or name in self._originally_adopted
        )
        if isinstance(replace, str) and replace in self._adopted:
            return ret
        if isinstance(replace, set) and replace & self._adopted:
            return ret
        return (
            ret
            and self._empire.emptyPolicySlots[policy.category] > 0
            and self._empire.policyPrereqsAndExclusionsOK(name)
        )

    def _get_adoptable(self) -> set[str]:
        """List of adoptable policies with the still available IP."""
        return {p for p in self._empire.availablePolicies if self._can_adopt(p)}

    def _adopt(self, name: str) -> None:
        """Find an emtpy slot and adopt named policy, if possible."""
        policy = fo.getPolicy(name)
        category = policy.category
        slot = self._find_empty_slot(category)
        # getting True does not guarantee that is has worked, so doublecheck the policy has been adopted
        if fo.issueAdoptPolicyOrder(name, category, slot) and name in self._empire.adoptedPolicies:
            if name not in self._originally_adopted:
                self._ip -= policy.adoptionCost()
            stats.adopt_policy(name)
            debug(f"Issued adoption order for {name} in slot {slot} turn {fo.currentTurn()}, remaining IP: {self._ip}")
            self._adoptable = self._get_adoptable()
            self._adopted.add(name)
        else:
            error("Failed to adopt %s in slot %d", name, slot)

    def _find_empty_slot(self, category: str) -> int:
        """This assumes there is an empty slot, otherwise adopt will fail."""
        slot = 0
        # there should be a better way...
        cat_slots = self._empire.categoriesSlotPolicies
        if category in cat_slots:
            slots = cat_slots[category]
            for slot in range(0, 99):
                if slot not in slots:
                    break
        return slot

    def _deadopt(self, name: str) -> None:
        """Deadopt name, if it is adopted."""
        if name in self._adopted:
            fo.issueDeadoptPolicyOrder(name)
            stats.deadopt_policy(name)
            self._adopted.remove(name)
            self._adoptable = self._get_adoptable()
            if name not in self._originally_adopted:
                # this is canceling an adoption order of the current turn
                self._ip += fo.getPolicy(name).adoptionCost()
            debug(f"Issued deadoption order for {name} turn {fo.currentTurn()}, remaining IP: {self._ip}")

    def _deadopt_one_of(self, names: set[str]) -> None:
        """
        Deadopt one of the given policies and store its name.
        This can be used after a decision is made, or to temporarily deadopt one for
        decision-making. In the later case, use _readopt_selected_one() to revert it.
        """
        selection = sorted((self._rate_policy(name), name) for name in names)
        debug(f"selection = {selection}")
        for _, name in selection:
            if name in self._adopted:
                self._selected_deadopt = name
                self._deadopt(name)
                break

    def _readopt_selected_one(self) -> None:
        """Re-adopt the policy deadopted by the last call to _deadopt_one_of()."""
        self._adopt(self._selected_deadopt)

    def _get_influence_priority(self) -> float:
        """Return current priority for resource type influence."""
        if fo.currentTurn() == 1:
            # AIState starts with 0, and we only set the value at the end of turn 1
            return initial_influence_priority
        else:
            return self._aistate.get_priority(PriorityType.RESOURCE_INFLUENCE)

    def print_status(self) -> None:
        # only for interactive debugging
        print("\nAdoptable Policies:")
        for p in self._get_adoptable():
            print("  ", p)
        print("Empty Slots:")
        for s in self._empire.emptyPolicySlots:
            print("  ", s)
        print("Adopted:")
        for category in self._empire.categoriesSlotPolicies:
            print(" ", category.key() + ":")
            for slot in category.data():
                print("   ", slot.key(), "->", slot.data(), "turn", self._empire.turnPolicyAdopted(slot.data()))
        print(f"Influence: {self._ip} (wanted: {self._wanted_ip})")
        print("Infl. Prod.:", self._get_infl_prod())
        print("Num Planet (pop./outp.): %d/%d" % (self._num_populated, self._num_outposts))


def generate_policy_orders() -> None:
    PolicyManager().generate_orders()


def print_status() -> None:
    PolicyManager(True).print_status()



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_INTSTEL_LOG",
    description="SHP_INTSTEL_LOG_DESC",
    short_description="STARLANE_SPEED_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=["CON_ORBITAL_CON"],
    unlock=[
        Item(type=UnlockPolicy, name="PLC_NO_SUPPLY"),
        Item(type=UnlockPolicy, name="PLC_CONFEDERATION"),
        Item(type=UnlockPolicy, name="PLC_TRAFFIC_CONTROL"),
    ],
    graphic="icons/tech/stellar_navigation.png",
)



from collections.abc import Mapping
from typing import Any

from common.print_utils._fields import Text
from common.print_utils._table import Table


def dict_to_table(data: Mapping[str, Any], name=None) -> Table:
    t = Table(Text("Key"), Text("Value", align=">"), hide_header=True, table_name=name)

    for key, val in data.items():
        t.add_row(key, str(val))
    return t



from common.statistic_interface import StatKey
from log_parser.log_tokenizer import _get_token_from_line


def test_token_parser():
    line = "12:06:13.179169 {0x00002e80} [debug] python : statistic.py:30 : ##EmpireID:empire_id: 2, name: Binding_2_pid_2_AI_1_RIdx_4_Aggressive, turn: 1"
    token = _get_token_from_line(line)
    assert len(token) == 2
    key, val = token
    assert key == StatKey.EmpireID
    assert val == "empire_id: 2, name: Binding_2_pid_2_AI_1_RIdx_4_Aggressive, turn: 1"



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import BROAD_EP, SWAMP_BROAD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import GOOD_HAPPINESS
from species.species_macros.industry import NO_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import GOOD_DEFENSE_TROOPS

Species(
    name="SP_CYNOS",
    description="SP_CYNOS_DESC",
    gameplay_description="SP_CYNOS_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    tags=[
        "PHOTOTROPHIC",
        "ARTISTIC",
        "NO_INDUSTRY",
        "GOOD_HAPPINESS",
        "AVERAGE_SUPPLY",
        "PEDIA_PHOTOTROPHIC_SPECIES_CLASS",
        "PEDIA_ARTISTIC",
    ],
    foci=[
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "SUCCULENT_BARNACLES_SPECIAL",
        "SHIMMER_SILK_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "ECCENTRIC_ORBIT_SPECIAL",
        "ANCIENT_RUINS_DEPLETED_SPECIAL",
        "FRUIT_SPECIAL",
        "PLC_DIVINE_AUTHORITY",
        "PLC_COLONIZATION",
        "PLC_ALLIED_REPAIR",
        "PLC_TERRAFORMING",
        "PLC_DESIGN_SIMPLICITY",
    ],
    dislikes=[
        "BLD_INDUSTRY_CENTER",
        "BLD_INTERSPECIES_ACADEMY",
        "BLD_ENCLAVE_VOID",
        "BLD_SOL_ORB_GEN",
        "BLD_PLANET_CLOAK",
        "PHILOSOPHER_SPECIAL",
        "RESONANT_MOON_SPECIAL",
        "TEMPORAL_ANOMALY_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "WORLDTREE_SPECIAL",
        "PLC_NO_GROWTH",
        "PLC_EXPLORATION",
        "PLC_NATIVE_APPROPRIATION",
    ],
    effectsgroups=[
        NO_INDUSTRY,
        *AVERAGE_RESEARCH,
        *AVERAGE_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *GOOD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_CYNOS"),
        *AVERAGE_SUPPLY,
        *GOOD_DEFENSE_TROOPS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        BROAD_EP,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=SWAMP_BROAD_EP,
    graphic="icons/species/cynos.png",
)



from focs._industry import _BASIC_INDUSTRY, _industry
from macros.misc import DESCRIPTION_EFFECTSGROUP_MACRO
from macros.multiplier import BAD_MULTIPLIER, GOOD_MULTIPLIER, GREAT_MULTIPLIER, VERY_BAD_MULTIPLIER

NO_INDUSTRY = DESCRIPTION_EFFECTSGROUP_MACRO("NO_INDUSTRY_DESC")

VERY_BAD_INDUSTRY = _industry("VERY_BAD", VERY_BAD_MULTIPLIER)
BAD_INDUSTRY = _industry("BAD", BAD_MULTIPLIER)
AVERAGE_INDUSTRY = _BASIC_INDUSTRY
GOOD_INDUSTRY = _industry("GOOD", GOOD_MULTIPLIER)
GREAT_INDUSTRY = _industry("GREAT", GREAT_MULTIPLIER)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_ASTEROID_HULLS",
    description="SHP_ASTEROID_HULLS_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=45 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_ASTEROID_HULL_TECHS"],
    prerequisites=["PRO_MICROGRAV_MAN"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_SMALL_ASTEROID"),
        Item(type=UnlockShipHull, name="SH_ASTEROID"),
        Item(type=UnlockBuilding, name="BLD_SHIPYARD_AST"),
    ],
    graphic="icons/ship_hulls/asteroid_hull_small.png",
)



from collections.abc import Mapping, Sequence, Set

import pytest

from common.fo_typing import PlanetId
from stub_generator.stub_generator.rtype.utils import get_name_for_mapping


@pytest.mark.parametrize(
    ("type_", "string"),
    [
        (int, "int"),
        (Sequence[int], "Sequence[int]"),
        (Sequence[PlanetId], "Sequence[PlanetId]"),
        (Set[int], "Set[int]"),
        (Set[PlanetId], "Set[PlanetId]"),
        (Mapping[int, PlanetId], "Mapping[int, PlanetId]"),
        (Sequence[Sequence[int]], "Sequence[Sequence[int]]"),
        (Sequence[Sequence[PlanetId]], "Sequence[Sequence[PlanetId]]"),
    ],
)
def test_get_name_for_mapping(type_, string):
    print()
    assert get_name_for_mapping(type_) == string



_nest_ratings = {
    "SNOWFLAKE_NEST_SPECIAL": 35.0,
    "KRAKEN_NEST_SPECIAL": 60.0,
    "JUGGERNAUT_NEST_SPECIAL": 80.0,
}


def special_is_nest(special_name: str) -> bool:
    return special_name in _nest_ratings


def get_nest_rating(special_name: str) -> float:
    return _nest_ratings.get(special_name, 0.0)



from focs._effects import (
    AllyOf,
    EffectsGroup,
    EmpireHasAdoptedPolicy,
    InSystem,
    IsSource,
    LocalCandidate,
    OwnedBy,
    ResourceSupplyConnected,
    SetStructure,
    Ship,
    Source,
    Stationary,
    Structure,
    Target,
    Turn,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_FLEET_REPAIR",
    description="SHP_FLEET_REPAIR_DESC",
    short_description="SHIP_REPAIR_DESC",
    category="SHIP_PARTS_CATEGORY",
    researchcost=130 * TECH_COST_MULTIPLIER,
    researchturns=10,
    tags=["PEDIA_DAMAGE_CONTROL_PART_TECHS"],
    prerequisites=["SHP_INTSTEL_LOG", "SHP_BASIC_DAM_CONT"],
    effectsgroups=[
        EffectsGroup(
            scope=Ship
            & InSystem()
            & Stationary
            & (
                OwnedBy(empire=Source.Owner)
                |
                # either ally providing repair or owner of ship beign repaired must adopt policy to share repairs
                (
                    EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_ALLIED_REPAIR")
                    | EmpireHasAdoptedPolicy(empire=LocalCandidate.Owner, name="PLC_ALLIED_REPAIR")
                )
                & OwnedBy(affiliation=AllyOf, empire=Source.Owner)
            )
            & Turn(low=LocalCandidate.System.LastTurnBattleHere + 1)
            & Structure(high=LocalCandidate.MaxStructure - 0.001)
            & ResourceSupplyConnected(empire=Source.Owner, condition=IsSource),
            stackinggroup="FLEET_REPAIR",
            effects=SetStructure(value=Value + (Target.MaxStructure / 10)),
        )
    ],
    graphic="icons/tech/fleet_field_repair.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="PRO_TRANS_DESIGN",
    description="PRO_TRANS_DESIGN_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["LRN_PSIONICS", "CON_ARCH_PSYCH"],
    unlock=Item(type=UnlockBuilding, name="BLD_INTERSPECIES_ACADEMY"),
    graphic="icons/tech/transcendent_architecture.png",
)



from io import StringIO

from interface_inspector.class_processor import ClassInfo

from stub_generator.stub_generator.class_generator import process_instantiated_class
from stub_generator.stub_generator.result_builder import Import, ResultBuilder, merge_imports, sort_imports


def test_merge_imports_same_imports_are_merged():
    imports = [
        Import("typing", ["Dict"]),
        Import("typing", ["List"]),
    ]
    result = merge_imports(imports)
    assert result == [Import("typing", ["Dict", "List"])]


def test_merge_imports_different_imports_are_not_merged():
    imports = [
        Import("collections", ["Dict"]),
        Import("typing", ["List"]),
    ]
    result = merge_imports(imports)
    assert result == imports


def test_sort_imports():
    imports = [
        Import("b", ["b", "a"]),
        Import("a", ["a"]),
    ]
    sort_imports(imports)
    assert imports == [
        Import("a", ["a"]),
        Import("b", ["a", "b"]),
    ]


expected_result = '''# header
from typing import Dict, List

from common.fo_typing import (
    BuildingId,
    EmpireId,
    XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX,
)

A = 1
B = 2

class Foo: ...

class Foo(Enum): ...

def foo(): ...
def boo():
    """doc"""

def goo():
    """doc"""
'''


def test_resource_builder():
    builder = ResultBuilder("# header")
    builder.add_built_in_import(Import("typing", ["List", "Dict"]))
    builder.add_import(Import("common.fo_typing", ["EmpireId", "BuildingId", "X" * 50]))
    builder.add_extra_declaration("\nA = 1")
    builder.add_extra_declaration("B = 2")
    builder.add_classes(["class Foo: ..."])
    builder.add_enums(["class Foo(Enum): ..."])
    builder.add_functions(["def foo(): ...", 'def boo():\n    """doc"""\n', 'def goo():\n    """doc"""\n'])

    res = StringIO()
    builder.write(res)
    res.seek(0)

    assert res.read() == expected_result


def test_process_instantiated_class():
    cls_info = ClassInfo(
        "diplomaticMessage",
        {
            "recipient": {"type": "<class 'property'>", "getter": None, "rtype": "int"},
            "sender": {"type": "<class 'property'>", "getter": None, "rtype": "int"},
            "type": {
                "type": "<class 'property'>",
                "getter": None,
                "rtype": "freeOrionAIInterface.diplomaticMessageType",
            },
        },
        None,
        [],
    )
    result = process_instantiated_class(cls_info)
    expected = "\n".join(
        [
            "class diplomaticMessage(NamedTuple):",
            "    recipient: PlayerId",
            "    sender: PlayerId",
            "    type: diplomaticMessageType",
        ]
    )

    assert result == expected



import pytest

from common.statistic_interface._serizlizer import (
    DictSerializer,
    to_float,
    to_int,
    to_str,
)


@pytest.mark.parametrize(
    ("converter", "serialized", "deserialized"),
    (
        (to_int, "11", 11),
        (to_float, "11.11", 11.11),
        (to_str, "hello", "hello"),
        (
            DictSerializer(
                {
                    "int": to_int,
                    "float": to_float,
                    "str": to_str,
                }
            ),
            "int: 11, float: 11.11, str: hello",
            {"int": 11, "float": 11.11, "str": "hello"},
        ),
    ),
)
def test_conversion(converter, serialized, deserialized):
    assert converter.deserialize(serialized) == deserialized



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import TOXIC_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import GOOD_HAPPINESS
from species.species_macros.industry import BAD_INDUSTRY
from species.species_macros.influence import BAD_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import GOOD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS, BAD_OFFENSE_TROOPS

Species(
    name="SP_GISGUFGTHRIM",
    description="SP_GISGUFGTHRIM_DESC",
    gameplay_description="SP_GISGUFGTHRIM_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    tags=[
        "ORGANIC",
        "BAD_INDUSTRY",
        "GOOD_RESEARCH",
        "BAD_INFLUENCE",
        "GOOD_HAPPINESS",
        "AVERAGE_SUPPLY",
        "BAD_OFFENSE_TROOPS",
        "PEDIA_ORGANIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "FRACTAL_GEODES_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "SUCCULENT_BARNACLES_SPECIAL",
        "MONOPOLE_SPECIAL",
        "ELERIUM_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "TEMPORAL_ANOMALY_SPECIAL",
        "PLC_CAPITAL_MARKETS",
        "PLC_ENVIRONMENTALISM",
        "PLC_TECHNOCRACY",
    ],
    dislikes=[
        "BLD_ENCLAVE_VOID",
        "BLD_PLANET_DRIVE",
        "BLD_PLANET_BEACON",
        "BLD_PLANET_CLOAK",
        "BLD_SPATIAL_DISTORT_GEN",
        "BLD_STARGATE",
        "BLD_HYPER_DAM",
        "KRAKEN_IN_THE_ICE_SPECIAL",
        "RESONANT_MOON_SPECIAL",
        "GAIA_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "PLC_INDUSTRIALISM",
        "PLC_NO_SUPPLY",
    ],
    effectsgroups=[
        *BAD_INDUSTRY,
        *GOOD_RESEARCH,
        *BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *GOOD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_GISGUFGTHRIM"),
        *AVERAGE_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *BAD_OFFENSE_TROOPS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=TOXIC_STANDARD_EP,
    graphic="icons/species/gis-guf-gthrim.png",
)



import freeOrionAIInterface as fo
from logging import warning
from typing import Optional

from AIDependencies import INVALID_ID
from common.fo_typing import FleetId, ObjectId, PlanetId, SystemId


class Target:
    """
    Stores information about AI target - its id and type.
    """

    object_name = "target"
    id: ObjectId

    def __init__(self):
        if not self:
            warning("Target is invalid %s" % self)

    def __eq__(self, other):
        return type(self) is type(other) and self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        target = self.get_object()
        if target is None:
            target = f"inaccessible {self.object_name}_{self.id}<>"
        return str(target)

    def get_object(self) -> Optional["fo.universeObject"]:
        return None

    def __bool__(self):
        return self.id is not None and self.id >= 0

    __nonzero__ = __bool__

    def get_system(self) -> Optional["TargetSystem"]:
        """
        Returns the system that contains this object, or None.
        """
        raise NotImplementedError()


class TargetPlanet(Target):
    object_name = "planet"

    def __init__(self, target_id: PlanetId):
        self.id: PlanetId = target_id
        super().__init__()

    def get_system(self) -> Optional["TargetSystem"]:
        """
        Returns the system that contains this object, or None.
        """
        universe = fo.getUniverse()
        planet = universe.getPlanet(self.id)
        return TargetSystem(planet.systemID)

    def get_object(self) -> "fo.planet":
        """
        :rtype fo.planet:
        """
        return fo.getUniverse().getPlanet(self.id)


class TargetSystem(Target):
    object_name = "system"

    def __init__(self, target_id: SystemId):
        self.id: SystemId = target_id
        super().__init__()

    def get_system(self) -> "TargetSystem":
        """
        Returns this object.
        """
        return self

    def get_object(self) -> "fo.system":
        return fo.getUniverse().getSystem(self.id)


class TargetFleet(Target):
    object_name = "fleet"

    def __init__(self, target_id: FleetId):
        self.id: FleetId = target_id
        super().__init__()

    def get_current_system_id(self):
        """
        Get current systemID (or INVALID_ID if on a starlane).
        """
        universe = fo.getUniverse()
        fleet = universe.getFleet(self.id)
        # will also return INVALID_ID if somehow the fleet cannot be retrieved
        return fleet.systemID if fleet else INVALID_ID

    def get_system(self) -> Optional[TargetSystem]:
        """
        Get current fleet location or target system if currently on starlane.
        """
        universe = fo.getUniverse()
        fleet = universe.getFleet(self.id)
        system_id = fleet.nextSystemID
        if system_id == INVALID_ID:  # fleet is not moving
            system_id = fleet.systemID
        return TargetSystem(system_id)

    def get_object(self) -> "fo.fleet":
        return fo.getUniverse().getFleet(self.id)



from focs._effects import (
    BlackHole,
    Contains,
    Destroy,
    EffectsGroup,
    GenerateSitRepMessage,
    IsBuilding,
    IsSource,
    Object,
    OwnedBy,
    Planet,
    Red,
    SetStarType,
    Source,
    Star,
    System,
)
from macros.base_prod import BUILDING_COST_MULTIPLIER
from macros.enqueue import ENQUEUE_BUILD_ONE_PER_PLANET

try:
    from focs._buildings import *
except ModuleNotFoundError:
    pass

BuildingType(  # type: ignore[reportUnboundVariable]
    name="BLD_ART_BLACK_HOLE",
    description="BLD_ART_BLACK_HOLE_DESC",
    buildcost=45 * BUILDING_COST_MULTIPLIER,
    buildtime=6,
    location=(
        Planet() & ~Contains(IsBuilding(name=["BLD_ART_BLACK_HOLE"])) & OwnedBy(empire=Source.Owner) & Star(type=[Red])
    ),
    enqueuelocation=ENQUEUE_BUILD_ONE_PER_PLANET,
    effectsgroups=[
        EffectsGroup(
            scope=Object(id=Source.SystemID) & System,
            activation=Star(type=[Red]),
            stackinggroup="ART_BLACK_HOLE",
            effects=[
                SetStarType(type=BlackHole),
                GenerateSitRepMessage(
                    message="EFFECT_BLACKHOLE",
                    label="EFFECT_BLACKHOLE_LABEL",
                    icon="icons/building/blackhole.png",
                    parameters={"system": Source.SystemID},
                    empire=Source.Owner,
                ),
            ],
        ),
        EffectsGroup(scope=IsSource, effects=Destroy),
    ],
    icon="icons/building/blackhole.png",
)



from focs._effects import (
    EffectsGroup,
    MaxOf,
    OwnedBy,
    Planet,
    SetDefense,
    SetMaxDefense,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.misc import PLANET_DEFENSE_FACTOR
from macros.priorities import DEFAULT_PRIORITY


def EG_DEFENSE_NET(multiplier: int, stack_postfix: str):
    return EffectsGroup(
        scope=Planet() & OwnedBy(empire=Source.Owner),
        stackinggroup="DEFENSE_NET_STACK_" + stack_postfix,
        accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
        priority=DEFAULT_PRIORITY,
        effects=SetMaxDefense(value=Value + (multiplier * PLANET_DEFENSE_FACTOR)),
    )


Tech(
    name="DEF_DEFENSE_NET_1",
    description="DEF_DEFENSE_NET_1_DESC",
    short_description="DEFENSE_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=30 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_ROOT_DEFENSE"],
    effectsgroups=[EG_DEFENSE_NET(5, "1")],
    graphic="icons/tech/defense.png",
)

Tech(
    name="DEF_DEFENSE_NET_2",
    description="DEF_DEFENSE_NET_2_DESC",
    short_description="DEFENSE_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=96 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_DEFENSE_NET_1"],
    effectsgroups=[EG_DEFENSE_NET(15, "2")],
    graphic="icons/tech/defense.png",
)

Tech(
    name="DEF_DEFENSE_NET_3",
    description="DEF_DEFENSE_NET_3_DESC",
    short_description="DEFENSE_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=240 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_DEFENSE_NET_2"],
    effectsgroups=[EG_DEFENSE_NET(25, "3")],
    graphic="icons/tech/defense.png",
)

Tech(
    name="DEF_DEFENSE_NET_REGEN_1",
    description="DEF_DEFENSE_NET_REGEN_1_DESC",
    short_description="DEFENSE_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_DEFENSE_NET_2"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            effects=SetDefense(value=Value + 0.1 * Target.MaxDefense),
        )
    ],
    graphic="icons/tech/defense_regeneration.png",
)

Tech(
    name="DEF_DEFENSE_NET_REGEN_2",
    description="DEF_DEFENSE_NET_REGEN_2_DESC",
    short_description="DEFENSE_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=240 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_DEFENSE_NET_REGEN_1"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            effects=SetDefense(value=Value + MaxOf(float, 0.1 * Target.MaxDefense, 0.25 * Target.Construction)),
        )
    ],
    graphic="icons/tech/defense_regeneration.png",
)
